# flake8: noqa
from ical_library.ical_components.v_alarm import VAlarm
from ical_library.ical_components.v_calendar import VCalendar
from ical_library.ical_components.v_event import VEvent, VRecurringEvent
from ical_library.ical_components.v_free_busy import VFreeBusy
from ical_library.ical_components.v_journal import VJournal, VRecurringJournal
from ical_library.ical_components.v_timezone import DayLight, Standard, VTimeZone
from ical_library.ical_components.v_todo import VRecurringToDo, VToDo



import heapq
from collections import defaultdict
from typing import Dict, Iterator, List, Optional, Tuple, Union

from pendulum import Date, DateTime

from ical_library.base_classes.component import Component
from ical_library.help_modules.lru_cache import instance_lru_cache
from ical_library.help_modules.timespan import Timespan, TimespanWithParent
from ical_library.ical_components import VCalendar, VFreeBusy
from ical_library.ical_components.abstract_components import AbstractComponentWithRecurringProperties


class Timeline:
    """
    This class is a wrapper to make it easy to see what the order of each component based on the start date.

    Inside this class there are multiple methods to iterate over all the components present. However, one should note
    that often the recurrence properties for components specify only a lower bound and will therefore proceed to
    infinity. To prevent us having an infinite list of items to iterate on, we define upper and lower bounds.
    You should set these upper bounds to the absolute minimum start date and absolute maximum end date that you would
    ever need. So if you need to do 10 different queries, this start and end date should range all of these, so it
    doesn't need to compute the list of components in the range over and over again. The functions themselves(e.g.
    :function:`self.includes` and :function:`self.intersects`) help you to limit the exact range you want to return
    components for.

    :param v_calendar: The VCalendar object we are iterating over.
    :param start_date: The minimum ending date of each event that is returned inside this timeline.
    :param end_date: The maximum starting date of each event that is return inside this timeline.
    """

    def __init__(
        self, v_calendar: VCalendar, start_date: Optional[DateTime] = None, end_date: Optional[DateTime] = None
    ):
        self.v_calendar: VCalendar = v_calendar
        self._start_date: DateTime = start_date or DateTime(1970, 1, 1)
        self._end_date: DateTime = end_date or DateTime(2100, 1, 1)

    def __repr__(self) -> str:
        return f"Timeline({self._start_date}, {self._end_date})"

    @property
    def start_date(self) -> DateTime:
        """Return the start date of the timeline. No event should end before this value."""
        return self._start_date

    @start_date.setter
    def start_date(self, value) -> None:
        """Set the start date of the timeline."""
        self._start_date = value
        self.get_timespan.cache_clear()

    @property
    def end_date(self) -> DateTime:
        """Return the end date of the timeline. No event should start before this value."""
        return self._end_date

    @end_date.setter
    def end_date(self, value) -> None:
        """Set the end date of the timeline."""
        self._end_date = value
        self.get_timespan.cache_clear()

    @instance_lru_cache()
    def get_timespan(self) -> Timespan:
        """Return the start and end date as a Timespan."""
        return Timespan(self.start_date, self.end_date)

    @staticmethod
    def __get_items_to_exclude_from_recurrence(
        all_components: List[Component],
    ) -> Dict[str, Union[List[Date], List[DateTime]]]:
        """
        Deduplicate recurring components. Sometimes it happens that recurring events are changed and this will cause
        them to both be present as a standard component and in the recurrence.
        :param all_components: The list of all component children of the VCalendar instance.
        :return: A deduplicated list of components.
        """
        start_date_to_timespan_dict: Dict[str, Union[List[Date], List[DateTime]]] = defaultdict(list)
        for component in all_components:
            if isinstance(component, AbstractComponentWithRecurringProperties) and component.recurrence_id is not None:
                start_date_to_timespan_dict[component.uid.value].append(component.recurrence_id.datetime_or_date_value)
        return start_date_to_timespan_dict

    def __explode_recurring_components(self) -> List[TimespanWithParent]:
        """
        Get a de-duplicated list of all components with a start date, including the recurring components. This means
        that we add all child component of the :class:`VCalendar` (except for the :class:`VTimeZone` instances) to a
        list and then add all extra occurrences (as recurring components) according to the recurrence properties:
        :class:`RRule`, :class:`RDate` and :class:`EXDate`.
        :return: A de-duplicated list of all components, including the recurring occurrences of the components.
        """
        list_of_timestamps_with_parents: List[TimespanWithParent] = []
        all_children = self.v_calendar.children
        uid_to_datetime_to_exclude = self.__get_items_to_exclude_from_recurrence(all_children)
        for c in all_children:
            # Do some initial filtering.
            if isinstance(c, AbstractComponentWithRecurringProperties):
                if c.max_recurring_timespan.intersects(self.get_timespan()):
                    values_to_exclude = uid_to_datetime_to_exclude[c.uid.value]  # .get defaults to None.
                    list_of_timestamps_with_parents.extend(
                        c.expand_component_in_range(self.get_timespan(), values_to_exclude)
                    )
            elif isinstance(c, VFreeBusy):
                if c.timespan.intersects(self.get_timespan()):
                    list_of_timestamps_with_parents.append(c.timespan)
            else:
                # There is no way to extend iana-props or x-props for now. If you feel like implementing this, please
                # let me know and open a PR :).
                pass
        return list_of_timestamps_with_parents

    def iterate(self) -> Iterator[Tuple[TimespanWithParent, Component]]:
        """
        Iterate over the `self.__explode_recurring_components()` in chronological order.

        Implementation detail: Using a heap is faster than sorting if the number of events (n) is much bigger than the
        number of events we extract from the iterator (k). Complexity: O(n + k log n).
        """
        heap: List[TimespanWithParent] = self.__explode_recurring_components()
        heapq.heapify(heap)
        while heap:
            popped: TimespanWithParent = heapq.heappop(heap)
            yield popped, popped.parent

    def includes(self, start: DateTime, stop: DateTime) -> Iterator[Component]:
        """Iterate (in chronological order) over every component that is in the specified timespan."""
        query_timespan = Timespan(start, stop)
        for timespan, event in self.iterate():
            if timespan.is_included_in(query_timespan):
                yield event

    def overlapping(self, start: DateTime, stop: DateTime) -> Iterator[Component]:
        """Iterate (in chronological order) over every component that has an intersection with the timespan."""
        query_timespan = Timespan(start, stop)
        for timespan, event in self.iterate():
            if timespan.intersects(query_timespan):
                yield event

    def start_after(self, instant: DateTime) -> Iterator[Component]:
        """Iterate (in chronological order) on every component larger than instant in chronological order."""
        for timespan, event in self.iterate():
            if timespan.begin > instant:
                yield event

    def at(self, instant: DateTime) -> Iterator[Component]:
        """Iterate (in chronological order) over all component that are occurring during `instant`."""
        for timespan, event in self.iterate():
            if timespan.includes(instant):
                yield event



from typing import List, Literal, Optional, Tuple, Union

import pendulum
from pendulum import Date, DateTime, Duration

from ical_library.base_classes.property import Property
from ical_library.help_modules import dt_utils


class _PeriodFunctionality(Property):
    """
    Provide methods to help to parse duration values.

    This class should be inherited by a Property.
    """

    def _parse_period_values(self) -> List[Tuple[DateTime, DateTime]]:
        """
        Parse multiple values, delimited by comma's, representing periods.

        Example value for self.value: 19960403T020000Z/19960403T040000Z,19960404T010000Z/PT3H
        :return: List of tuples containing two DateTimes representing the start and end of the duration respectively.
        """
        list_of_periods: List[Tuple[DateTime, DateTime]] = []
        for item in self.value.split(","):
            item = item.strip()
            instance = self._parse_individual_duration_str(item)
            if not isinstance(instance, tuple) or len(instance) != 2:
                raise TypeError(f"{instance} is of {type(instance)=} while it should be a tuple.")
            for index, sub_instance in enumerate(instance):
                if not isinstance(sub_instance, DateTime):
                    raise TypeError(
                        f"{instance[index]=} is of {type(sub_instance)=} while it should be of type "
                        f"Tuple[DateTime, DateTime]."
                    )
            list_of_periods.append(instance)
        return list_of_periods

    def _parse_individual_datetime_or_duration_str(self, datetime_or_duration_str: str) -> Union[DateTime, Duration]:
        """
        Parse an individual datetime or duration string.
        :param datetime_or_duration_str: A string represent either a datetime or duration.
        :return: A pendulum.DateTime if the string represented a datetime. Return a pendulum.Duration otherwise.
        """
        tz_id = self.get_property_parameter("TZID")
        tz = self.parent.tree_root.get_timezone(tz_id).get_as_timezone_object() if tz_id else None
        return pendulum.parse(datetime_or_duration_str, tz=tz)

    def _parse_individual_duration_str(self, period_str: str) -> Tuple[DateTime, DateTime]:
        """
        Parse an individual period represented by DateTime/DateTime or DateTime/Duration.

        :param period_str: The period to parse. Examples: 19960403T020000Z/19960403T040000Z or 19960404T010000Z/PT3H
        :return: A tuple containing two DateTimes representing the start and end of the duration respectively.
        """
        first_str, second_str = period_str.split("/")
        first_instance: DateTime = self._parse_individual_datetime_or_duration_str(first_str)
        second_instance: Union[DateTime, Duration] = self._parse_individual_datetime_or_duration_str(second_str)
        if not isinstance(first_instance, DateTime):
            raise TypeError(f"Expected {period_str=} to contain a DateTime as first argument.")
        if isinstance(second_instance, DateTime):
            return first_instance, second_instance
        elif isinstance(second_instance, Duration):
            computed_datetime: DateTime = first_instance + second_instance  # type: ignore  # Pendulum at fault
            return first_instance, computed_datetime
        else:
            raise TypeError(f"Expected {period_str=} to contain a DateTime or Duration as second argument.")


class _ExOrRDate(_PeriodFunctionality):
    """
    Provide methods to help to parse different kind of values a Property could find.

    This class should be inherited by a Property.
    """

    def _parse_datetime_values(self) -> List[DateTime]:
        """
        Parses DateTime values. Example of a possible value of self.value: 19970714T123000Z,19970714T123300Z
        :return: A List of DateTimes representing the start of an event/component.
        """
        list_of_datetimes: List[DateTime] = []
        for item in self.value.split(","):
            item = item.strip()
            instance = self._parse_individual_datetime_or_duration_str(item)
            if not isinstance(instance, DateTime):
                raise TypeError(f"{instance} is of {type(instance)=} while it should be a DateTime.")
            list_of_datetimes.append(instance)
        return list_of_datetimes

    def _parse_date_values(self) -> List[Date]:
        """
        Parse Date values. Example of a possible value of self.value: 19970101,19970120,19970217
        :return: A list of pendulum.Date representing the start of an event/component.
        """
        list_of_dates: List[Date] = []
        for item in self.value.split(","):
            instance = self._parse_individual_date_str(item)
            if not isinstance(instance, Date):
                raise TypeError(f"{instance} is of {type(instance)=} while it should be a Date.")
            list_of_dates.append(instance)
        return list_of_dates

    @staticmethod
    def _parse_individual_date_str(date: str) -> Date:
        """
        Parse an individual date string.
        :param date: A string representing a date.
        :return: A pendulum.Date.
        """
        return Date(int(date[0:4]), int(date[4:6]), int(date[6:8]))


class FreeBusyProperty(_PeriodFunctionality):
    """
    The FREEBUSY property defines one or more free or busy time intervals.

    Note: This class is called FreeBusyProperty to not be confused with the VFreeBusy component.
    """

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *FREEBUSYPROPERTY* but *FREEBUSY*."""
        return "FREEBUSY"

    @property
    def free_busy_type(self) -> str:
        """
        Specifies the free or busy time type.

        Values are usually in the following list but can be anything: FREE, BUSY, BUSY-UNAVAILABLE, BUSY-TENTATIVE
        """
        return self.get_property_parameter_default("FBTYPE", "BUSY")

    @property
    def periods(self) -> List[Tuple[DateTime, DateTime]]:
        """
        All the periods present in this property for which we define a free or busy time.
        :return: A list of tuples, where each tuple values consists of two DateTimes indicating the start and end
        respectively.
        """
        return self._parse_period_values()


class EXDate(_ExOrRDate):
    """
    The EXDATE property defines the list of DATE-TIME exceptions for recurring events, to-dos, journal entries,
    or time zone definitions.
    """

    @property
    def kind(self) -> Optional[Literal["DATE-TIME", "DATE"]]:
        """The kind of the values. It is either DATE-TIME or DATE. The default is DATE-TIME."""
        return self.property_parameters.get("VALUE", "DATE-TIME")

    @property
    def excluded_date_times(self) -> Union[List[DateTime], List[Date]]:
        """A list of all excluded Dates or DateTimes. The type will be according to kind reported by `self.kind()`."""
        if self.kind == "DATE-TIME":
            return self._parse_datetime_values()
        elif self.kind == "DATE":
            return self._parse_date_values()
        else:
            raise ValueError(f"{self.kind=} should be one in ['DATE-TIME', 'DATE'].")


class RDate(_ExOrRDate):
    """
    The RDATE property defines the list of DATE-TIME values for recurring events, to-dos, journal entries,
    or time zone definitions.
    """

    @property
    def kind(self) -> Optional[Literal["DATE-TIME", "DATE", "PERIOD"]]:
        """The kind of the values. It is either DATE-TIME, DATE or PERIOD. The default is DATE-TIME."""
        return self.property_parameters.get("VALUE", "DATE-TIME")

    @property
    def all_values(self) -> Union[List[DateTime], List[Date], List[Tuple[DateTime, DateTime]]]:
        """
        A list of all recurring Dates, DateTimes or Periods. The periods are defined by tuples containing two
        datetimes representing the start and stop respectively. The returned types in the list will be according to
        the kind reported by `self.kind()`.
        """
        if self.kind == "DATE-TIME":
            return self._parse_datetime_values()
        elif self.kind == "DATE":
            return self._parse_date_values()
        elif self.kind == "PERIOD":
            return self._parse_period_values()
        else:
            raise ValueError(f"{self.kind=} should be one in ['DATE-TIME', 'DATE', 'PERIOD'].")

    def compute_max_end_date(self, component_duration: Duration) -> DateTime:
        """
        To speed up the computation of the Timelines range, it's good to know the ending of the last recurring event
        of a recurrence property. This does not need to be perfect, it should just be an estimate (so we don't check
        EXDate and such).
        :param component_duration: The duration of the component which has the recurring properties.
        :return: An estimate of the maximum end date across all occurrences. This value should always be at least the
        actual highest recurrence end date
        """
        max_value: Optional[DateTime] = None
        for value in self.all_values:
            if isinstance(value, Date):  # This covers both Date and DateTime
                dt: DateTime = dt_utils.convert_time_object_to_datetime(value)
                dt: DateTime = dt + component_duration  # type: ignore  # Pendulum at fault here.
                dt = dt_utils.convert_time_object_to_aware_datetime(dt)
                if max_value is None or dt > max_value:
                    max_value = dt
            elif isinstance(value, tuple) and len(value) == 2:
                dt: DateTime = dt_utils.convert_time_object_to_aware_datetime(value[1])
                if max_value is None or dt > max_value:
                    max_value = dt
            else:
                raise ValueError(f"Unexpected value encountered: {value}.")
        return max_value or DateTime.max



from typing import Iterator, List, Optional, Union

from pendulum import Date, DateTime, Duration

from ical_library.base_classes.component import Component
from ical_library.help_modules import property_utils
from ical_library.help_modules.timespan import Timespan, TimespanWithParent
from ical_library.ical_components.abstract_components import (
    AbstractComponentWithRecurringProperties,
    AbstractRecurrence,
)
from ical_library.ical_properties.cal_address import Attendee, Organizer
from ical_library.ical_properties.dt import _DTBoth, Created, DTStamp, DTStart, LastModified, RecurrenceID
from ical_library.ical_properties.ints import Sequence
from ical_library.ical_properties.pass_properties import (
    Attach,
    Categories,
    Class,
    Comment,
    Contact,
    Description,
    RelatedTo,
    RequestStatus,
    Status,
    Summary,
    UID,
    URL,
)
from ical_library.ical_properties.periods import EXDate, RDate
from ical_library.ical_properties.rrule import RRule


class VJournal(AbstractComponentWithRecurringProperties):
    """
    This class represents the VJOURNAL component specified in RFC 5545 in '3.6.3. Journal Component'.

    A "VJOURNAL" calendar component is a grouping of component properties that represent one or more descriptive text
    notes associated with a particular calendar date. The "DTSTART" property is used to specify the calendar date with
    which the journal entry is associated. Generally, it will have a DATE value data type, but it can also be used to
    specify a DATE-TIME value data type. Examples of a journal entry include a daily record of a legislative body or
    a journal entry of individual telephone contacts for the day or an ordered list of accomplishments for the day.
    The "VJOURNAL" calendar component can also be used to associate a document with a calendar date.

    :param name: The actual name of this component instance. E.g. VEVENT, RRULE, VCUSTOMCOMPONENT.
    :param dtstamp: The DTStamp property. Required and must occur exactly once.
    :param uid: The UID property. Required and must occur exactly once.
    :param dtstart: The DTStart property. Optional and may occur at most once.
    :param rrule: The RRule property. Optional and may occur at most once.
    :param summary: The Summary property. Optional and may occur at most once.
    :param exdate: The EXDate property. Optional, but may occur multiple times.
    :param rdate: The RDate property. Optional, but may occur multiple times.
    :param comment: The Comment property. Optional, but may occur multiple times.
    :param ical_class: Optional Class property. Optional, but may occur at most once.
    :param created: The Created property. Optional, but may occur at most once.
    :param last_modified: Optional LastModified property. Optional, but may occur at most once.
    :param organizer: The Organizer property. Optional, but may occur at most once.
    :param sequence: The Sequence property. Optional, but may occur at most once.
    :param status: The Status property. Optional, but may occur at most once.
    :param url: The URL property. Optional, but may occur at most once.
    :param recurrence_id: Optional RecurrenceID property. Optional, but may occur at most once.
    :param attach: The Attach property. Optional, but may occur multiple times.
    :param attendee: The Attendee property. Optional, but may occur multiple times.
    :param categories: The Categories property. Optional, but may occur multiple times.
    :param contact: The Contact property. Optional, but may occur multiple times.
    :param description: The Description property. Optional, but may occur multiple times.
    :param related: The RelatedTo property. Optional, but may occur multiple times.
    :param rstatus: The RequestStatus property. Optional, but may occur multiple times.
    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    """

    def __init__(
        self,
        dtstamp: Optional[DTStamp] = None,
        uid: Optional[UID] = None,
        dtstart: Optional[DTStart] = None,
        rrule: Optional[RRule] = None,
        summary: Optional[Summary] = None,
        exdate: Optional[List[EXDate]] = None,
        rdate: Optional[List[RDate]] = None,
        comment: Optional[List[Comment]] = None,
        ical_class: Optional[Class] = None,
        created: Optional[Created] = None,
        last_modified: Optional[LastModified] = None,
        organizer: Optional[Organizer] = None,
        sequence: Optional[Sequence] = None,
        status: Optional[Status] = None,
        url: Optional[URL] = None,
        recurrence_id: Optional[RecurrenceID] = None,
        attach: Optional[List[Attach]] = None,
        attendee: Optional[List[Attendee]] = None,
        categories: Optional[List[Categories]] = None,
        contact: Optional[List[Contact]] = None,
        description: Optional[List[Description]] = None,
        related: Optional[List[RelatedTo]] = None,
        rstatus: Optional[List[RequestStatus]] = None,
        parent: Optional[Component] = None,
    ):
        super().__init__(
            name="VJOURNAL",
            dtstamp=dtstamp,
            uid=uid,
            dtstart=dtstart,
            rrule=rrule,
            summary=summary,
            exdate=exdate,
            rdate=rdate,
            comment=comment,
            parent=parent,
        )

        # Optional, may only occur once
        # As class is a reserved keyword in python, we prefixed it with `ical_`.
        self.ical_class: Optional[Class] = self.as_parent(ical_class)
        self.created: Optional[Created] = self.as_parent(created)
        self.last_modified: Optional[LastModified] = self.as_parent(last_modified)
        self.organizer: Optional[Organizer] = self.as_parent(organizer)
        self.sequence: Optional[Sequence] = self.as_parent(sequence)
        self.status: Optional[Status] = self.as_parent(status)
        self.url: Optional[URL] = self.as_parent(url)

        # Optional, may occur more than once
        self.attach: Optional[List[Attach]] = self.as_parent(attach)
        self.attendee: Optional[List[Attendee]] = self.as_parent(attendee)
        self.categories: Optional[List[Categories]] = self.as_parent(categories)
        self.contact: Optional[List[Contact]] = self.as_parent(contact)
        self.description: Optional[List[Description]] = self.as_parent(description)
        self.related: Optional[List[RelatedTo]] = self.as_parent(related)
        self.rstatus: Optional[List[RequestStatus]] = self.as_parent(rstatus)

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"VJournal({self.start}: {self.summary.value if self.summary else ''})"

    @property
    def ending(self) -> Optional[_DTBoth]:
        """
        Return the start time of the journal. This is because the Journal does not have a duration.

        Note: This is an abstract method from :class:`AbstractComponentWithRecurringProperties` we have to implement.
        """
        return self.dtstart

    def get_duration(self) -> Optional[Duration]:
        """
        Return an empty Duration as a Journal does not have a duration.

        Note: This is an abstract method from :class:`AbstractComponentWithRecurringProperties` we have to implement.
        """
        return Duration()

    def expand_component_in_range(
        self, return_range: Timespan, starts_to_exclude: Union[List[Date], List[DateTime]]
    ) -> Iterator[TimespanWithParent]:
        """
        Expand this VJournal in range according to its recurring *RDate*, *EXDate* and *RRule* properties.
        :param return_range: The timespan range on which we should return VJournal instances.
        :param starts_to_exclude: List of start Dates or list of start DateTimes of which we already know we should
        exclude them from our recurrence computation (as they have been completely redefined in another element).
        :return: Yield all recurring VJournal instances related to this VJournal in the given *return_range*.
        """
        if self.timespan.intersects(return_range):
            yield self.timespan
        starts_to_exclude.append(self.start)

        start = self.start
        duration = self.computed_duration
        if not start or not duration:
            return None

        iterator = property_utils.expand_component_in_range(
            exdate_list=self.exdate or [],
            rdate_list=self.rdate or [],
            rrule=self.rrule,
            first_event_start=self.start,
            first_event_duration=self.computed_duration,
            starts_to_exclude=starts_to_exclude,
            return_range=return_range,
            make_tz_aware=None,
        )

        for event_start_time, event_end_time in iterator:
            yield VRecurringJournal(original_component_instance=self, start=event_start_time).timespan


class VRecurringJournal(AbstractRecurrence, VJournal):
    """
    This class represents VJournal that are recurring.
    Inside the AbstractRecurrence class we overwrite specific dunder methods and property methods.
    This way our end users have a very similar interface to an actual VJournal but without us needing to code the exact
    same thing twice.

    :param original_component_instance: The original VJournal instance.
    :param start: The start of this occurrence.
    :param end: The end of this occurrence.
    """

    def __init__(self, original_component_instance: VJournal, start: DateTime):
        self._original = original_component_instance
        self._start = start
        self._end = start
        super(VJournal, self).__init__("VJOURNAL", parent=original_component_instance)

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"RVJournal({self._start}: {self.original.summary.value if self.original.summary else ''})"



import os
from pathlib import Path

import pytest

from ical_library import client
from ical_library.ical_components import VCalendar
from ical_library.ical_properties import ProdID


@pytest.fixture
def root_folder() -> Path:
    starting_place = Path(os.getcwd()).resolve()
    folder = starting_place
    while not (folder / ".git").is_dir() and not folder.name.lower() == "iCal-library":
        folder = folder.parent
        if folder.name == "":
            raise ValueError(f"Could not find the root folder starting from {starting_place=}.")
    return folder.resolve()


@pytest.fixture
def calendar_instance() -> VCalendar:
    return VCalendar(
        prodid=ProdID("-//Google Inc//Google Calendar 70.9054//EN"),
    )


@pytest.fixture
def calendar_with_all_components_once(root_folder: Path) -> VCalendar:
    return client.parse_icalendar_file(root_folder / "tests" / "resources" / "iCalendar-with-all-components-once.ics")


@pytest.fixture
def calendar_with_reoccurring_events_once(root_folder: Path) -> VCalendar:
    return client.parse_icalendar_file(root_folder / "tests" / "resources" / "iCalendar-with-reoccurring-events.ics")


@pytest.fixture
def empty_calendar(root_folder: Path) -> VCalendar:
    return client.parse_icalendar_file(root_folder / "tests" / "resources" / "iCalender-without-anything.ics")


@pytest.fixture
def berlin_timezone_calendar(root_folder: Path) -> VCalendar:
    return client.parse_icalendar_file(root_folder / "tests" / "resources" / "iCalendar-with-berlin-timezone.ics")


@pytest.fixture
def multi_offset_calendar(root_folder: Path) -> VCalendar:
    filename = "iCalendar-with-recurring-event-multiple-timezone-offsets.ics"
    return client.parse_icalendar_file(root_folder / "tests" / "resources" / filename)


@pytest.fixture
def calendar_exdate(root_folder: Path) -> VCalendar:
    return client.parse_icalendar_file(root_folder / "tests" / "resources" / "iCalendar-exdate.ics")



from ical_library.ical_properties.cal_address import _CalAddress


def test_cal_address():
    ca = _CalAddress(name="ORGANIZER", property_parameters="CN=John Smith", value="mailto:jsmith@example.com")
    assert ca.email == "jsmith@example.com"
    assert ca.persons_name == "John Smith"



from typing import Tuple

from ical_library.base_classes.property import Property


class GEO(Property):
    """
    The GEO property specifies information related to the global position for the activity specified by a calendar
    component.
    """

    @property
    def geo_value(self) -> Tuple[float, float]:
        """Return the value as two floats representing the latitude and longitude."""
        latitude, longitude = self.value.split(";")
        return float(latitude), float(longitude)



from typing import Dict, Iterator, List, Literal, Optional, Tuple, Union

import pendulum
from dateutil import rrule as base_for_time_periods
from dateutil.rrule import rrule, weekday
from pendulum import Date, DateTime, Duration

from ical_library.base_classes.property import Property
from ical_library.help_modules import dt_utils
from ical_library.help_modules.lru_cache import instance_lru_cache


class RRule(Property):
    """
    The RRULE property defines a rule or repeating pattern for recurring events, to-dos, journal entries, or time zone
    definitions.

    For more in depth restrictions and possibilities we refer you to the `RTFC 5545` section `3.3.10. Recurrence Rule`.
    """

    @property
    @instance_lru_cache()
    def value_as_dict(self) -> Dict[str, str]:
        """
        Parse all recurrence rule parts as a dictionary, so it can be used as an easy lookup.
        :return: A dict mapping of str to str containing all recurrence rule parts.
        """
        all_values = [tuple(key_and_value.split("=")) for key_and_value in self._value.split(";")]
        return {key: value for key, value in all_values}

    @property
    def freq(self) -> str:
        """
        The FREQUENCY rule identifies the type of recurrence rule. Possible values are: SECONDLY, MINUTELY, HOURLY,
        DAILY, WEEKLY, MONTHLY and YEARLY.
        This is the only required field.
        :return: The frequency as a string.
        """
        return self.value_as_dict["FREQ"]

    @property
    def freq_dateutil(self) -> int:
        """
        Return the frequency in the format dateutil expects which is a map of the string to an integer.
        :return: An integer in the range of 0 to 6.
        """
        return getattr(base_for_time_periods, self.freq)

    @property
    def until(self) -> Optional[Union[Date, DateTime]]:
        """
        The UNTIL rule defines a DATE or DATE-TIME value that bounds the recurrence rule in an inclusive manner.
        This is optional but may not occur together with COUNT.
        :return: None or a positive integer.
        """
        if (until_value := self.value_as_dict.get("UNTIL", None)) is not None:
            return dt_utils.parse_date_or_datetime(until_value)
        return None

    @property
    def count(self) -> Optional[int]:
        """
        The COUNT rule defines the number of occurrences at which to range-bound the recurrence.
        This is optional but may not occur together with UNTIL.
        :return: None or a positive integer.
        """
        count_value = self.value_as_dict.get("COUNT", None)
        return int(count_value) if count_value else None

    @property
    def interval(self) -> int:
        """
        The INTERVAL rule contains a positive integer representing at which intervals the recurrence rule repeats.
        :return: A positive integer.
        """
        return int(self.value_as_dict.get("INTERVAL", 1))

    @staticmethod
    def convert_str_to_optional_integer_tuple(value: Optional[str]) -> Optional[Tuple[int, ...]]:
        """
        Converts a string to a Tuple of integers.
        :return: None or a Tuple of integers if the value exists, otherwise None.
        """
        if not value:
            return None
        return tuple(int(item) for item in value.split(","))

    @property
    def by_second(self) -> Optional[Tuple[int, ...]]:
        """
        The BYSECOND rule part specifies a COMMA-separated list of seconds within a minute.
        :return: None or a list of integers in the range of 0 to 60.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYSECOND"))

    @property
    def by_minute(self) -> Optional[Tuple[int, ...]]:
        """
        The BYMINUTE rule part specifies a COMMA-separated list of minutes within an hour.
        :return: None or a list of integers in the range of 0 to 59.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYMINUTE"))

    @property
    def by_hour(self) -> Optional[Tuple[int, ...]]:
        """
        The BYHOUR rule part specifies a COMMA-separated list of hours of the day.
        :return: None or a list of integers in the range of 0 to 23.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYHOUR"))

    def by_day(self) -> Optional[List[Tuple[Optional[int], Literal["SU", "MO", "TU", "WE", "TH", "FR", "SA"]]]]:
        """
        The BYDAY rule part specifies a COMMA-separated list of days of the week;
        SU indicates Sunday; MO indicates Monday; TU indicates Tuesday; WE indicates Wednesday;
        TH indicates Thursday; FR indicates Friday; and SA indicates Saturday.

        Each BYDAY value can also be preceded by a positive (+n) or negative (-n) integer.
        If present, this indicates the nth occurrence of a specific day within the MONTHLY or YEARLY "RRULE".

        Example values are SU,TU or +2SU,-3TU

        :return None or a list of tuples of two values. The first value represents possible specified nth occurrence or
        None. The second value the day of the week as a SU, MO, TU, WE, TH, FR or SA.
        """
        value = self.value_as_dict.get("BYDAY")
        if not value:
            return None
        list_of_days: List[Tuple[Optional[int], Literal["SU", "MO", "TU", "WE", "TH", "FR", "SA"]]] = []
        for a_day in value.split(","):
            a_day = a_day.strip()
            nth_occurence: Optional[int] = int(a_day[:-2]) if len(a_day) > 2 else None
            day_of_week: str = a_day[-2:]
            if day_of_week not in ("SU", "MO", "TU", "WE", "TH", "FR", "SA"):
                raise ValueError
            list_of_days.append((nth_occurence, day_of_week))  # type: ignore
        return list_of_days

    @property
    def by_day_dateutil(self) -> Optional[Tuple[weekday, ...]]:
        """
        Return the by_day in the format dateutil expects which is a tuple of weekday instance.
        :return: None or a tuple of weekday instances which is a type native to dateutil.
        """
        day_list: List[weekday] = []
        by_day = self.by_day()
        if by_day is None:
            return None
        for optional_nth_occurrence, weekday_str in by_day:
            try:
                if optional_nth_occurrence is not None:
                    day_list.append(getattr(base_for_time_periods, weekday_str)(optional_nth_occurrence))
                else:
                    day_list.append(getattr(base_for_time_periods, weekday_str))
            except Exception as exc:
                raise ValueError(f"{optional_nth_occurrence=}, {weekday_str=}") from exc
        return tuple(day_list)

    @property
    def by_month_day(self) -> Optional[Tuple[int, ...]]:
        """
        The BYMONTHDAY rule part specifies a COMMA-separated list of days of the month.
        For example: -10 represents the tenth to the last day of the month.
        :return: None or a tuple of integers in the range of 1 to 31 or -31 to -1.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYMONTHDAY"))

    @property
    def by_year_day(self) -> Optional[Tuple[int, ...]]:
        """
        The BYYEARDAY rule part specifies a COMMA-separated list of days of the year.
        For example: -1 represents the last day of the year (December 31st).
        :return: None or a tuple of integers in the range of 1 to 366 or -366 to -1.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYYEARDAY"))

    @property
    def by_week_no(self) -> Optional[Tuple[int, ...]]:
        """
        The BYWEEKNO rule part specifies a COMMA-separated list of ordinals specifying weeks of the year.
        For example: 3 represents the third week of the year.
        :return: None or an integer in the range of 1 to 53 or -53 to -1.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYWEEKNO"))

    @property
    def by_month(self) -> Optional[Tuple[int, ...]]:
        """
        The BYMONTH rule part specifies a COMMA-separated list of months of the year.
        :return: None or a list of integers in the range of 1 to 12.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYMONTH"))

    @property
    def by_set_pos(self) -> Optional[Tuple[int, ...]]:
        """
        The BYSETPOS rule part specifies a COMMA-separated list of values that corresponds to the nth occurrence within
        the set of recurrence instances specified by the rule. BYSETPOS operates on a set of recurrence instances in
        one interval of the recurrence rule. For example, in a WEEKLY rule, the interval would be one week A set of
        recurrence instances starts at the beginning of the interval defined by the FREQ rule part.
        :return: None or a list of integers in the range of 1 to 366 or -366 to -1.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYSETPOS"))

    @property
    def wkst(self) -> Optional[str]:
        """
        The WKST rule part specifies the day on which the workweek starts.
        :return: A string that is the value of MO, TU, WE, TH, FR, SA, or SU or None.
        """
        day = self.value_as_dict.get("WKST")
        if not day:
            return None
        if day not in ("SU", "MO", "TU", "WE", "TH", "FR", "SA"):
            raise ValueError(f"{day=} is not in the list of weekdays.")
        return day

    @property
    def wkst_dateutil(self) -> Optional[int]:
        """
        Return the wkst in the format dateutil expects which is an integer. 0 for MO, 1 for TU, 2 for WE, ...
        :return: An integer in the range of 0 to 6 or None.
        """
        day = self.wkst
        return getattr(base_for_time_periods, day) if self.wkst is not None else None

    @property
    def by_easter(self) -> Optional[Tuple[int, ...]]:
        """
        According to dateutil, this is an extension of the RFC specification.. I can't find it. If anyone can, please
        file an issue or a PR to add it here as a reference.

        The BYEASTER rule part specifies the offset from the Easter Sunday.
        :return: None or an integer in the range of 1 to 366 or -366 to -1 or None.
        """
        return self.convert_str_to_optional_integer_tuple(self.value_as_dict.get("BYEASTER"))

    def compute_max_end_date(self, starting_datetime: Union[Date, DateTime], component_duration: Duration) -> DateTime:
        """
        To speed up the computation of the Timelines range, it's good to know the ending of the last recurring event
        of a recurrence property. This does not need to be perfect, it should just be an estimate (so we don't check
        EXDate and such).
        :param starting_datetime: The starting datetime from which we start computing the next occurrences.
        :param component_duration: The duration of the component which has the recurring properties.
        :return: An estimate of the maximum end date across all occurrences. This value should always be at least the
        actual highest recurrence end date
        """
        if self.until:
            return dt_utils.convert_time_object_to_aware_datetime(self.until) + component_duration  # type: ignore
        elif self.count:
            if self.count < 1000:
                max_datetime = DateTime.max if isinstance(starting_datetime, DateTime) else Date(9999, 12, 31)
                *_, last = self.sequence_iterator(starting_datetime=starting_datetime, max_datetime=max_datetime)
                return dt_utils.convert_time_object_to_aware_datetime(last) + component_duration  # type: ignore
        return DateTime.max

    def sequence_iterator(
        self, starting_datetime: Union[Date, DateTime], max_datetime: Union[Date, DateTime]
    ) -> Iterator[DateTime]:
        """
        Given a starting datetime, we compute dates according to the RRule specification until the end of the sequence
        according to the specification is reached or until we reached the max_datetime.
        :param starting_datetime: The starting datetime from which we start computing the next occurrences.
        :param max_datetime: The maximum datetime. If we reach this datetime, we stop the iteration..
        :return: Yield all datetimes(except itself) in the sequence.
        """
        if type(starting_datetime) != type(max_datetime):
            raise TypeError(f"{type(starting_datetime)=} and {type(max_datetime)=} should be of the same type.")
        if isinstance(starting_datetime, DateTime):
            if (starting_datetime.tz or max_datetime.tz) and (not starting_datetime.tz or not max_datetime.tz):
                raise TypeError(f"The tz info should be consistent: {starting_datetime=}, {max_datetime=}.")
        if not isinstance(starting_datetime, (Date, DateTime)) or not isinstance(max_datetime, (Date, DateTime)):
            raise TypeError(f"{type(starting_datetime)=} and {type(max_datetime)=} should be of Date or DateTime.")
        if starting_datetime >= max_datetime:
            raise ValueError(f"This should not be the case: {starting_datetime=} >= {max_datetime=} .")

        if not (is_datetime_format := isinstance(starting_datetime, DateTime) and isinstance(max_datetime, DateTime)):
            starting_datetime = DateTime(starting_datetime.year, starting_datetime.month, starting_datetime.day)
            max_datetime = DateTime(max_datetime.year, max_datetime.month, max_datetime.day)

        starting_datetime = starting_datetime
        max_datetime = max_datetime
        starting_tz = starting_datetime.tz
        until = self.until
        if until:
            if not isinstance(until, DateTime):
                until = DateTime(until.year, until.month, until.day)
            until = until.replace(tzinfo=None) if starting_tz is None else until.in_timezone(starting_tz)

        keyword_arguments = {
            "until": until,
            "count": self.count,
            "interval": self.interval,
            "bysecond": self.by_second,
            "byminute": self.by_minute,
            "byhour": self.by_hour,
            "byweekday": self.by_day_dateutil,
            "bymonthday": self.by_month_day,
            "byyearday": self.by_year_day,
            "byweekno": self.by_week_no,
            "bymonth": self.by_month,
            "bysetpos": self.by_set_pos,
            "wkst": self.wkst_dateutil,
            "byeaster": self.by_easter,
        }
        no_none_keywords = {key: value for key, value in keyword_arguments.items() if value is not None}
        dt_iterator = rrule(dtstart=starting_datetime, freq=self.freq_dateutil, **no_none_keywords)
        for dt in dt_iterator:
            if dt > max_datetime:
                break
            p_instance = pendulum.instance(dt, tz=None)
            yield p_instance if is_datetime_format else p_instance.date()



from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ical_library import Component


class ComponentContext:
    """
    Component context is used to keep the current Component when Component is used as ContextManager.

    You can use components as context: `#!py3 with Component() as my_component:`.

    If you do this the context stores the Component and whenever a new component or property is created, it will use
    such stored Component as the parent Component.
    """

    _context_managed_component: Optional["Component"] = None
    _previous_context_managed_components: List["Component"] = []

    @classmethod
    def push_context_managed_component(cls, component: "Component"):
        """Set the current context managed component."""
        if cls._context_managed_component:
            cls._previous_context_managed_components.append(cls._context_managed_component)
        cls._context_managed_component = component

    @classmethod
    def pop_context_managed_component(cls) -> Optional["Component"]:
        """Pop the current context managed component."""
        old_component = cls._context_managed_component
        if cls._previous_context_managed_components:
            cls._context_managed_component = cls._previous_context_managed_components.pop()
        else:
            cls._context_managed_component = None
        return old_component

    @classmethod
    def get_current_component(cls) -> Optional["Component"]:
        """Get the current context managed component."""
        return cls._context_managed_component



# flake8: noqa
from ical_library.ical_properties.cal_address import Attendee, Organizer
from ical_library.ical_properties.dt import Completed, Created, DTEnd, DTStamp, DTStart, Due, LastModified, RecurrenceID
from ical_library.ical_properties.geo import GEO
from ical_library.ical_properties.ical_duration import ICALDuration
from ical_library.ical_properties.ints import PercentComplete, Priority, Repeat, Sequence
from ical_library.ical_properties.pass_properties import (
    Action,
    Attach,
    CalScale,
    Categories,
    Class,
    Comment,
    Contact,
    Description,
    Location,
    Method,
    ProdID,
    RelatedTo,
    RequestStatus,
    Resources,
    Status,
    Summary,
    TimeTransparency,
    TZID,
    TZName,
    TZURL,
    UID,
    URL,
    Version,
)
from ical_library.ical_properties.periods import EXDate, FreeBusyProperty, RDate
from ical_library.ical_properties.rrule import RRule
from ical_library.ical_properties.trigger import Trigger
from ical_library.ical_properties.tz_offset import TZOffsetFrom, TZOffsetTo



from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ical_library.base_classes.component import Component


class ICalBaseClass:
    """
    This is the base class of all custom classes representing an iCal component or iCal property in our library.

    [ical_library.base_classes.Component][] and :class:`Property` are the only ones inheriting this class directly, the
    rest of the classes are inheriting from :class:`Component` and :class:`Property` based on whether they represent an
    iCal component or iCal property.

    :param name: the actual name of this property or component. E.g. VEVENT, RRULE, VCUSTOMCOMPONENT, CUSTOMPROPERTY.
    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    """

    def __init__(self, name: str, parent: Optional["Component"]):
        self._name = name
        if self._name is None:
            raise ValueError("Name of a Component or Property should not be None. Please specify it.")
        self._parent: Optional["Component"] = parent

    @property
    def parent(self) -> Optional["Component"]:
        """
        Return the parent :class:`Component` that contains this :class:`Component`.
        :return: Return the parent :class:`Component` instance or None in the case there is no parent (for VCalender's).
        """
        return self._parent

    @parent.setter
    def parent(self, value: "Component"):
        """
        Setter for the parent :class:`Component`. This allows us to set the parent at a later moment.
        :param value: The parent :class:`Component`.
        """
        self._parent = value

    @property
    def name(self) -> str:
        """
        Return the actual name of this property or component. E.g. VEVENT, RRULE, VCUSTOMCOMPONENT, CUSTOMPROPERTY.

        We inherit this class, for the general Property and Component but also for the specific VEvent component and
        the RRule property. Now what do we do with the `x-comp` or `iana-comp` components and `x-prop` and `iana-prop`
        properties? They also have an iCalendar name, e.g. VCUSTOMCOMPONENT. However, we can't specify them beforehand
        as we simply can't cover all cases. Therefore, we use `get_ical_name_of_class` to find and map all of our
        pre-defined Components and Properties but we still specify the name for all custom components. So the rule of
        thumb:
        Use `.name` on instantiated classes while we use `.get_ical_name_of_class()` for non-instantiated classes.
        """
        return self._name

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """
        Return the name of a pre-defined property or pre-defined component. E.g. VEVENT, RRULE, COMPONENT, PROPERTY.

        For a :class:`Property` this would be the value at the start of the line. Example: a property with the name of
        `ABC;def=ghi:jkl` would be `ABC`.
        For a :class:`Component` this would be the value at the start of the component after BEGIN. Example: a VEvent
        starts with `BEGIN:VEVENT`, hence this function would return `VEVENT`.
        """
        return cls.__name__.upper()



import inspect
import re
from collections import defaultdict
from functools import lru_cache
from typing import (
    Callable,
    ClassVar,
    Dict,
    get_args,
    get_origin,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from ical_library.base_classes.base_class import ICalBaseClass
from ical_library.base_classes.property import Property
from ical_library.exceptions import CalendarParentRelationError
from ical_library.help_modules.component_context import ComponentContext

if TYPE_CHECKING:
    from ical_library.ical_components import VCalendar

T = TypeVar("T")


class Component(ICalBaseClass):
    """
    This is the base class for any component (according to the RFC 5545 specification) in iCal-library.

    Inside all components (so also all classes inheriting this class, e.g. VEvent) there are four kind of variables:

    - variables that start with `_`. These are metadata of the class and not parsed as a property or
     component from the iCalendar data file.
    - variables that have a type of `List[x]` and a default value of List. These are child components/properties of
    the instance. These components/properties may or may not be required to be present in the iCal file.
    - variables that have a type of `Optional[List[x]]`. These are components/properties of the instance.
    They can be either optional or required and may occur multiple times in the iCal file.
    - variables that have a type of `Optional[x]` (and not `Optional[List[x]]`). These are properties of the instance.
    They can be either optional or required, but may only occur once in the iCal file.

    Any Component that is predefined according to the RFC 5545 should inherit this class, e.g. VCalendar, VEVENT.
    Only x-components or iana-components should instantiate the Component class directly.

    :param name: The actual name of this component instance. E.g. `VEVENT`, `RRULE`, `VCUSTOMCOMPONENT`.
    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    """

    def __init__(self, name: str, parent: Optional["Component"] = None):
        name = name if self.__class__ == Component else self.__class__.get_ical_name_of_class()
        super().__init__(name=name, parent=parent or ComponentContext.get_current_component())
        if parent is None and self.parent is not None:
            self.parent.__add_child_without_setting_the_parent(self)
        if self.parent is None and self.name != "VCALENDAR":
            raise ValueError("We should always set a Parent on __init__ of Components.")
        self._parse_line_start: Optional[int] = 0
        self._parse_line_end: Optional[int] = 0
        self._extra_child_components: Dict[str, List["Component"]] = defaultdict(list)
        self._extra_properties: Dict[str, List[Property]] = defaultdict(list)

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        properties_as_string = ", ".join([f"{name}={value}" for name, value in self.properties.items()])
        return f"{self.__class__.__name__}({properties_as_string})"

    def __eq__(self: "Component", other: "Component") -> bool:
        """Return whether the current instance and the other instance are the same."""
        if type(self) != type(other):
            return False
        return self.properties == other.properties and self.children == other.children

    def __enter__(self):
        """Enter the context manager. Check ComponentContext for more info."""
        ComponentContext.push_context_managed_component(self)
        return self

    def __exit__(self, _type, _value, _tb):
        """Exit the context manager. Check ComponentContext for more info."""
        ComponentContext.pop_context_managed_component()

    def _set_self_as_parent_for_ical_component(self, prop_or_comp: ICalBaseClass) -> None:
        """Verifies the parent is not already set to a different component, if not sets the parent."""
        if prop_or_comp.parent is None:
            prop_or_comp.parent = self
        elif prop_or_comp.parent != self:
            raise ValueError(
                "Trying to overwrite a parent. Please do not re-use property instance across different components."
            )

    def as_parent(self, value: T) -> T:
        """We set self as Parent for Properties and Components but also Properties and Components in lists."""
        if isinstance(value, ICalBaseClass):
            self._set_self_as_parent_for_ical_component(value)
        elif isinstance(value, list):  # checking for list over Iterable is ~8,5x faster.
            for item in value:
                if isinstance(item, ICalBaseClass):
                    self._set_self_as_parent_for_ical_component(item)
        return value

    @property
    def extra_child_components(self) -> Dict[str, List["Component"]]:
        """Return all children components that are considered as `x-comp` or `iana-comp` components."""
        return self._extra_child_components

    @property
    def extra_properties(self) -> Dict[str, List[Property]]:
        """Return all properties that are considered as `x-prop` or `iana-prop` properties."""
        return self._extra_properties

    @property
    def tree_root(self) -> "VCalendar":
        """Return the tree root which should always be a VCalendar object."""
        from ical_library.ical_components import VCalendar

        instance = self
        while instance.parent is not None:
            instance = instance.parent
        if not isinstance(instance, VCalendar):
            raise CalendarParentRelationError(
                f"TreeRoot {instance=} is of type {type(instance)=} instead of VCalendar."
            )
        return instance

    @property
    def children(self) -> List["Component"]:
        """Return all children components."""
        extras = (child for list_of_children in self._extra_child_components.values() for child in list_of_children)
        children = [
            item_in_list
            for ical_name, (var_name, var_type, is_list) in self._get_child_component_mapping().items()
            for item_in_list in getattr(self, var_name)
        ]
        children.extend(extras)
        return children

    def __add_child_without_setting_the_parent(self, child: "Component") -> None:
        """
        Just add a child component and do not also set the parent.

        If the child is an undefined `x-comp` or `iana-comp` component, we add it to _extra_child_components.
        If the child is defined, we add it to one of the other variables according to
        :function:`self._get_child_component_mapping()`.
        """
        child_component_mapping = self._get_child_component_mapping()
        if child.name in child_component_mapping:
            var_name, var_type, is_list = child_component_mapping[child.name]
            getattr(self, var_name).append(child)
            return
        self._extra_child_components[child.name].append(child)

    def add_child(self, child: "Component") -> None:
        """
        Add a children component and set its parent.

        If the child is an undefined `x-comp` or `iana-comp` component, we add it to _extra_child_components.
        If the child is defined, we add it to one of the other variables according to
        :function:`self._get_child_component_mapping()`.
        """
        self.as_parent(child)
        self.__add_child_without_setting_the_parent(child)

    @property
    def original_ical_text(self) -> str:
        """Return the original iCAL text for your property from the RAW string list as if it is a property."""
        return self.tree_root.get_original_ical_text(self._parse_line_start, self._parse_line_end)

    @classmethod
    @lru_cache()
    def get_property_ical_names(cls) -> Set[str]:
        """
        Get all the variables for this component class that reference a :class:`Property` in the typing information.
        """
        return {var_name for var_name, var_type, is_list in cls._get_property_mapping().values()}

    @staticmethod
    def _extract_ical_class_from_args(var_name: str, a_type: Union[Type[List], type(Union)]) -> Type:
        """
        Given *a_type*, which is either a List or an Optional, return the subtype that is not None.

        Note: When we execute get_args(some_type), we consider the result to be the subtypes.
        :param var_name: The variable name of the type we are dissecting.
        :param a_type: The type we want to get the subtype of.
        :return: The subtype that is not equal to the NoneType.
        :raise: TypeError when there is no subtype that does not contain a type that is not equal to NoneType.
        """
        sub_types: List[Type] = [st for st in get_args(a_type) if not issubclass(get_origin(st) or st, type(None))]
        if len(sub_types) != 1:
            raise TypeError(f"Incorrect number of sub_types to follow here for {var_name=}, {a_type=}, {sub_types=}.")
        return sub_types[0]

    @staticmethod
    def _extract_type_information(
        var_name: str, a_type: Type, is_in_list: bool
    ) -> Optional[Tuple[str, Tuple[str, Optional[Type[ICalBaseClass]], bool]]]:
        """
        Extract typing information for an instance variable of the component.

        The type of the variable should either be (wrapping) a :class:`Property` or a :class:`Component`.
        :param var_name: The variable name of the type we are dissecting.
        :param a_type: The type we want to extract a child class of :class:`ICalBaseClass` from.
        :param is_in_list: Whether the child class of :class:`ICalBaseClass` is contained in a List type.
        :return: None if there is no child class of :class:`ICalBaseClass` we can detect. Otherwise, we return
        a tuple containing the iCal name (e.g. VEVENT) and another tuple that contains the variable name, the child
        class of :class:`ICalBaseClass` and a boolean whether that child class was wrapped in a List.
        :raise: TypeError if there is no child class of :class:`ICalBaseClass` to detect.
        """
        if get_origin(a_type) is None:
            if issubclass(a_type, ICalBaseClass):
                return a_type.get_ical_name_of_class(), (var_name, a_type, is_in_list)
            return None
        elif get_origin(a_type) == Union:  # This also covers the Optional case.
            sub_class = Component._extract_ical_class_from_args(var_name, a_type)
            return Component._extract_type_information(var_name, sub_class, is_in_list)
        elif issubclass(get_origin(a_type), List):
            sub_class = Component._extract_ical_class_from_args(var_name, a_type)
            return Component._extract_type_information(var_name, sub_class, True)
        elif get_origin(a_type) == ClassVar:
            return None
        else:
            raise TypeError(f"Unknown type '{a_type}' came by in Component.extract_custom_type.")

    @classmethod
    def _get_init_method_for_var_mapping(cls) -> Callable:
        """
        We generate _get_var_mapping based on `cls.__init__`. This var mapping is later used to list all properties,
        all components but also all the types of the items. This is a function so that it can be overwritten for
        the recurring components.
        """
        return cls.__init__

    @classmethod
    @lru_cache()
    def _get_var_mapping(cls) -> Mapping[str, Tuple[str, Type[ICalBaseClass], bool]]:
        """
        Get a mapping of all variables of this class that do not start with `_`.
        :return: A class mapping that maps the iCal name (e.g. VEVENT) to another tuple that contains
        the variable name, the child class of :class:`ICalBaseClass` and a boolean whether that child class was wrapped
        in a List.
        """
        var_mapping: Dict[str, Tuple[str, Type[ICalBaseClass], bool]] = {}
        a_field: inspect.Parameter
        for a_field in inspect.signature(cls._get_init_method_for_var_mapping()).parameters.values():
            if a_field.name.startswith("_") or a_field.name in ["self", "parent", "name"]:
                continue
            result = Component._extract_type_information(a_field.name, a_field.annotation, False)
            if result is None:
                continue
            ical_name, var_type_info = result
            if issubclass(var_type_info[1], ICalBaseClass):
                var_mapping[ical_name] = var_type_info
        return var_mapping

    @classmethod
    @lru_cache()
    def _get_property_mapping(cls) -> Mapping[str, Tuple[str, Type[Property], bool]]:
        """
        Return the same mapping as :function:`cls._get_var_mapping()` but only return variables related to
        :class:`Property` classes. Example: `{"RRULE": tuple("rrule", Type[RRule], False), ...}`
        """
        return {
            ical_name: var_tuple
            for ical_name, var_tuple in cls._get_var_mapping().items()
            if issubclass(var_tuple[1], Property)
        }

    @classmethod
    @lru_cache()
    def _get_child_component_mapping(cls) -> Mapping[str, Tuple[str, Type["Component"], bool]]:
        """
        Return the same mapping as :function:`cls._get_var_mapping()` but only return variables related to
        :class:`Component` classes.
        """
        return {
            ical_name: var_tuple
            for ical_name, var_tuple in cls._get_var_mapping().items()
            if issubclass(var_tuple[1], Component)
        }

    @property
    def properties(self) -> Dict[str, Union[Property, List[Property]]]:
        """Return all iCalendar properties of this component instance."""
        standard_properties = {
            var_name: getattr(self, var_name)
            for var_name, var_type, is_list in self._get_property_mapping().values()
            if getattr(self, var_name) is not None
        }
        return {**standard_properties, **self._extra_properties}

    def print_tree_structure(self, indent: int = 0) -> None:
        """Print the tree structure of all components starting with this instance."""
        print(f"{'  ' * indent} - {self}")
        for child in self.children:
            child.print_tree_structure(indent=indent + 1)

    def set_property(
        self,
        property_instance: Property,
        property_map_info: Optional[Tuple[str, Type[Property], bool]] = None,
        property_map_was_checked: bool = False,
    ) -> None:
        """
        Setting a property for a Component instance.

        If the `property_map_info` is equal to None, we either have not yet looked up all the properties or it is an
        x-prop/iana-prop. This can be decided based on property_map_was_checked. This avoids extra (expensive) lookups
        in our self._get_property_mapping.

        :param property_instance: The Property we wish to set.
        :param property_map_info: A tuple containing the variable name for this Component instance, the type of the
        Property and whether the variable can occur multiple times for the same property. If it equals None, this either
        means it is an x-prop/iana-prop or that the property_map was not checked yet.
        :param property_map_was_checked: Whether the `property_map_info` was passed or not. If True, and
        `property_map_info` is `None`, we know that it is an iana property. If False, we still need to consult
        `_get_property_mapping`.
        """
        if property_map_info is None and property_map_was_checked is False:
            property_map_info = self._get_property_mapping().get(property_instance.name)
        var_name, var_type, is_list = property_map_info or [None, None, None]
        if var_name is not None and is_list is not None:
            if is_list is True:
                if getattr(self, var_name) is None:
                    setattr(self, var_name, [property_instance])
                else:
                    current_value: List[Property] = getattr(self, var_name)
                    current_value.append(property_instance)
            else:
                setattr(self, var_name, property_instance)
        else:
            self._extra_properties[property_instance.name.lower().replace("-", "_")].append(property_instance)

    def parse_property(self, line: str) -> Property:
        """
        Parse a raw line containing a :class:`Property` definition, instantiate the corresponding Property and set the
        variable.

        Based on the first part of the line (before the ; and :), we know using *self._get_property_mapping()* which
        property type we should instantiate. Then, depending on whether the typing info of the property denoted it in
        a List or not, it adds it to a list/instantiates the list, compared to simply setting it as the variable of
        the :class:`Component` instance.

        Credits for the excellent regex parsing string go to @Jan Goyvaerts: https://stackoverflow.com/a/2482067/2277445

        :param line: The entire line that contains the property string (meaning multi-lines properties are already
        converted to a single line here).
        :return: The created Property instance based on the *line* that we set for the component.
        """
        property_mapping = self._get_property_mapping()
        result = re.search("^([^\r\n;:]+)(;[^\r\n:]+)?:(.*)$", line)
        if result is None:
            raise ValueError(f"{result=} should never be None! {line=} is invalid.")
        name, property_parameters, value = result.group(1), result.group(2), result.group(3)
        property_parameters = property_parameters.lstrip(";") if property_parameters else None
        if name in property_mapping.keys():
            property_map_info = property_mapping[name]
            var_name, var_type, is_list = property_map_info
            property_instance = var_type(name=name, property_parameters=property_parameters, value=value, parent=self)
            self.set_property(property_instance, property_map_info, property_map_was_checked=True)
        else:
            property_instance = Property(name=name, property_parameters=property_parameters, value=value, parent=self)
            self.set_property(property_instance, None, property_map_was_checked=True)
        return property_instance

    def _instantiate_component(self, ical_component_identifier: str) -> "Component":
        component_mapping = self._get_child_component_mapping()
        if ical_component_identifier in component_mapping:
            var_name, var_type, is_list = component_mapping[ical_component_identifier]
            return var_type(parent=self)  # type: ignore
        else:
            return Component(name=ical_component_identifier, parent=self)

    def parse_component(self, lines: List[str], line_number: int) -> int:
        """
        Parse the raw lines representing this component (which was just instantiated).

        Based on the first line that starts with `BEGIN:`, we know using *self._get_child_component_mapping()* which
        specific component type we should instantiate. We then add it to the current component instance as a child.
        Then we parse line by line, if we find another `BEGIN:`, we create another component instance and proceed
        to calling :function:`self.parse_component` for parsing all the lines related to that component. If we find
        a property line (any line that doesn't start with `BEGIN:`), we call :function:`self.parse_property` which
        then automatically adds it to the current instance.

        :param lines: A list of all the lines in the iCalendar file.
        :param line_number: The line number at which this component starts.
        :return: The line number at which this component ends.
        """
        self._parse_line_start = line_number - 1
        while not (current_line := lines[line_number]).startswith("END:"):
            line_number += 1
            if current_line.startswith("BEGIN:"):
                component_name = current_line[len("BEGIN:") :]
                instance = self._instantiate_component(component_name)
                self.add_child(instance)
                line_number = instance.parse_component(lines=lines, line_number=line_number)
                continue

            full_line_without_line_breaks = current_line
            while (next_line := lines[line_number]).startswith(" "):
                line_number += 1
                # [1:] so we skip the space indicating a line break.
                full_line_without_line_breaks += next_line[1:]
            self.parse_property(full_line_without_line_breaks)

        if current_line != f"END:{self.name}":
            raise ValueError(
                f"Expected {current_line=} to be equal to END:{self.name}. It seems {self} was never closed."
            )
        self._parse_line_end = line_number + 1
        return line_number + 1



from typing import Union

import pendulum
from pendulum import Date, DateTime
from pendulum.tz.timezone import FixedTimezone

from ical_library.base_classes.property import Property
from ical_library.help_modules import dt_utils


class _DTBoth(Property):
    """This property class should be inherited. It represents a property that contain a datetime or date as value."""

    @property
    def datetime_or_date_value(self) -> Union[Date, DateTime]:
        """Return the value as a DateTime or Date value taking into account the optional TZID property parameter."""
        value = dt_utils.parse_date_or_datetime(self.value)
        if isinstance(value, DateTime):
            tz_id = self.get_property_parameter("TZID")
            if value.tz or not tz_id:
                return value
            return self.parent.tree_root.get_aware_dt_for_timezone(dt=value, tzid=tz_id)
        elif isinstance(value, Date):
            return value
        else:
            raise TypeError(f"Unknown {type(value)=} returned for {value=}.")

    def get_datetime_or_date_value_in_specific_tz(self, tz: FixedTimezone) -> Union[Date, DateTime]:
        """Return the value as a DateTime or Date value in a specific timezone."""
        value = dt_utils.parse_date_or_datetime(self.value)
        if isinstance(value, DateTime):
            return value.in_timezone(tz)
        elif isinstance(value, Date):
            return value
        else:
            raise TypeError(f"Unknown {type(value)=} returned for {value=}.")


class _DTSingular(Property):
    """This property class should be inherited. It represents a property that can only contain a datetime as value."""

    @property
    def datetime(self) -> DateTime:
        """Return the value as a DateTime value taking into account the optional TZID property parameter."""
        value = pendulum.parse(self.value, tz=None)
        tz_id = self.get_property_parameter("TZID")
        if value.tz or not tz_id:
            return value
        return self.parent.tree_root.get_aware_dt_for_timezone(dt=value, tzid=tz_id)


# Value & TZInfo
class DTStart(_DTBoth):
    """The DTSTART property specifies when the calendar component begins.."""

    pass


# Value & TZInfo
class DTEnd(_DTBoth):
    """The DTEND property specifies the date and time that a calendar component ends."""

    pass


# Value & TZInfo
class Due(_DTBoth):
    """This DUE property defines the date and time that a to-do is expected to be completed.."""

    pass


# Value & TZInfo
class RecurrenceID(_DTBoth):
    """
    The RECURRENCE-ID property is defined as followed.

    This property is used in conjunction with the "UID" and "SEQUENCE" properties to identify a specific instance of
    a recurring "VEVENT", "VTODO", or "VJOURNAL" calendar component. The property value is the original value of the
    "DTSTART" property of the recurrence instance. Value Type: The default value type is DATE-TIME. The value type
    can be set to a DATE value type. This property MUST have the same value type as the "DTSTART" property contained
    within the recurring component. Furthermore, this property MUST be specified as a date with local time if and
    only if the "DTSTART" property contained within the recurring component is specified as a date with local time.
    """

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *LASTMODIFIED* but *LAST-MODIFIED*."""
        return "RECURRENCE-ID"


# Only date-time
class DTStamp(_DTBoth):
    """The DTSTAMP property is defined as followed.

    In the case of an iCalendar object that specifies a "METHOD" property, this property specifies the date and time
    that the instance of the iCalendar object was created. In the case of an iCalendar object that doesn't specify a
    "METHOD" property, this property specifies the date and time that the information associated with the calendar
    component was last revised in the calendar store.
    """

    pass


# Only date-time
class Completed(_DTSingular):
    """The COMPLETED property defines the date and time that a to-do was actually completed."""

    pass


# Only date-time
class Created(_DTSingular):
    """
    The CREATED property defines the date and time that the calendar information was created by the calendar
    user agent in the calendar store.
    """

    pass


# Only date-time
class LastModified(_DTSingular):
    """
    The LAST-MODIFIED property specifies the date and time that the information associated with the calendar component
    was last revised in the calendar store.
    """

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *LASTMODIFIED* but *LAST-MODIFIED*."""
        return "LAST-MODIFIED"



import pendulum

from ical_library.base_classes.property import Property


class ICALDuration(Property):
    """The DURATION property specifies a positive duration of time."""

    @property
    def duration(self) -> pendulum.Duration:
        """Return the value as a parsed pendulum.Duration. Example value: PT1H0M0S."""
        parsed_value: pendulum.Duration = pendulum.parse(self.value)
        if not isinstance(parsed_value, pendulum.Duration):
            raise TypeError(f"Invalid value passed for Duration: {self.value=}")
        return parsed_value

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *ICALDURATION* but *DURATION*."""
        return "DURATION"



import pytest

from ical_library.ical_components import (
    DayLight,
    Standard,
    VAlarm,
    VCalendar,
    VEvent,
    VFreeBusy,
    VJournal,
    VTimeZone,
    VToDo,
)


def test_get_child_component_mapping_with_children():
    assert VCalendar._get_child_component_mapping() == {
        "VEVENT": ("events", VEvent, True),
        "VFREEBUSY": ("free_busy_list", VFreeBusy, True),
        "VJOURNAL": ("journals", VJournal, True),
        "VTIMEZONE": ("time_zones", VTimeZone, True),
        "VTODO": ("todos", VToDo, True),
    }

    assert VTimeZone._get_child_component_mapping() == {
        "DAYLIGHT": ("daylightc", DayLight, True),
        "STANDARD": ("standardc", Standard, True),
    }

    assert VEvent._get_child_component_mapping() == {
        "VALARM": ("alarms", VAlarm, True),
    }

    assert VToDo._get_child_component_mapping() == {
        "VALARM": ("alarms", VAlarm, True),
    }


@pytest.mark.parametrize("a_type", [VFreeBusy, VJournal, DayLight, Standard, VAlarm])
def test_get_child_component_mapping_without_children(a_type):
    assert a_type._get_child_component_mapping() == {}



import pendulum
from pendulum import DateTime

from ical_library.ical_properties.dt import DTStart


def test_dt_both(berlin_timezone_calendar):
    with berlin_timezone_calendar:
        a = DTStart(name="DTSTART", property_parameters="TZID=Europe/Berlin", value="20190307T020000")
    assert a.datetime_or_date_value == DateTime(2019, 3, 7, 2, 0, 0, tzinfo=pendulum.timezone("Europe/Berlin"))



from ical_library.base_classes.component import Component
from ical_library.base_classes.property import Property
from ical_library.ical_components import VCalendar, VEvent, VFreeBusy, VJournal, VTimeZone, VToDo
from ical_library.ical_properties.pass_properties import ProdID, Version


def test_repr(calendar_instance):
    with calendar_instance:
        my_component = Component("MORTHY")
        my_component._extra_properties["prop"].append(Property(name="PROP", property_parameters=None, value="GHI"))
        assert repr(my_component) == "Component(prop=[Property(PROP:GHI)])"


def test_extra_child_components(calendar_instance):
    with calendar_instance:
        with Component("MORTHY", None) as my_component:
            component_1 = Component("RICK", None)
            component_2 = Component("SUMMER", None)
    assert my_component.extra_child_components == {
        "RICK": [component_1],
        "SUMMER": [component_2],
    }


def test_extra_properties(calendar_instance):
    my_component = Component("MORTHY", parent=calendar_instance)
    my_component.parse_property("RICK:SUMMER")
    assert set(my_component.properties.keys()) == {"rick"}
    print(dir(my_component.properties.values()))
    property = list(my_component.properties.values())[0][0]
    assert property.name == "RICK"
    assert property.value == "SUMMER"


def test_parent(calendar_instance):
    some_component = Component("SOME-COMPONENT", calendar_instance)
    assert some_component.parent is calendar_instance
    with calendar_instance:
        a = Property(name="PROP", value="bcdef")
        assert a.parent is calendar_instance


def test_children(calendar_with_all_components_once: VCalendar):
    assert len(calendar_with_all_components_once.children) == 6
    type_of_children = [type(child) for child in calendar_with_all_components_once.children]
    assert type_of_children == [VEvent, VToDo, VJournal, VFreeBusy, VTimeZone, Component]


def test_original_ical_text(calendar_with_all_components_once: VCalendar):
    assert (
        calendar_with_all_components_once.free_busy_list[0].original_ical_text
        == """
BEGIN:VFREEBUSY
UID:19970901T082949Z-FA43EF@example.com
ORGANIZER:mailto:jane_doe@example.com
ATTENDEE:mailto:john_public@example.com
DTSTART:19971015T050000Z
DTEND:19971016T050000Z
DTSTAMP:19970901T083000Z
END:VFREEBUSY
    """.strip()
    )


def test_properties(empty_calendar):
    version_property = Version(name="VERSION", value="1.1")
    empty_calendar.version = version_property
    some_other_property = Property(name="AWESOME", value="VALUE")
    empty_calendar._extra_properties["awesome"].append(some_other_property)
    assert empty_calendar.properties["awesome"] == [some_other_property]
    assert empty_calendar.properties["version"] == version_property
    assert set(empty_calendar.properties.keys()) == {
        "prodid",
        "version",
        "calscale",
        "method",
        "x_wr_caldesc",
        "x_wr_timezone",
        "x_wr_calname",
        "awesome",
    }


def test_print_tree_structure(capsys):
    root = VCalendar(prodid=ProdID(name="A", value="B"), version=Version(name="C", value="D"))
    an_event = VEvent(parent=root)
    a_journal = VJournal(parent=root)
    root.add_child(an_event)
    root.add_child(a_journal)
    root.print_tree_structure()
    captured = capsys.readouterr()
    assert captured.out == " - VCalendar(B, D)\n   - VEvent()\n   - VJournal(None: )\n"



from typing import Optional, Union

import pendulum
from pendulum import Date, DateTime
from pendulum.tz import get_local_timezone
from pendulum.tz.timezone import Timezone


def parse_date_or_datetime(value: str) -> Union[Date, DateTime]:
    """Parse a string into a pendulum.Date or pendulum.Datetime."""
    if len(value) == 8:
        return pendulum.Date(int(value[0:4]), int(value[4:6]), int(value[6:8]))
    return pendulum.parse(value, tz=None)


def convert_time_object_to_datetime(time_value: Union[Date, DateTime]) -> DateTime:
    """Convert the argument to a pendulum.DateTime object whether it's a pendulum.Date or a pendulum.DateTime."""
    if isinstance(time_value, DateTime):
        return time_value
    return DateTime(time_value.year, time_value.month, time_value.day)


def make_datetime_aware(dt: DateTime, tz: Optional[Union[str, Timezone]] = None):
    """
    Make a pendulum.DateTime timezone aware.

    When it already has a timezone, it doesn't change the timezone. If it doesn't, it will add the timezone.
    """
    if dt.tz is not None:
        return dt
    timezone: Timezone = (tz if isinstance(tz, Timezone) else pendulum.timezone(tz)) if tz else get_local_timezone()
    return dt.in_timezone(timezone)


def convert_time_object_to_aware_datetime(
    time_value: Union[Date, DateTime], tz: Optional[Union[str, Timezone]] = None
) -> DateTime:
    """Convert a time pendulum.Date or pendulum.DateTime to a timezone aware pendulum.DateTime."""
    dt: DateTime = convert_time_object_to_datetime(time_value)
    return make_datetime_aware(dt, tz)



from ical_library.base_classes.property import Property


class ProdID(Property):
    """The PRODID property specifies the identifier for the product that created the iCalendar object."""

    pass


class Version(Property):
    """
    The VERSION property specifies the identifier corresponding to the highest version number or the minimum and
    maximum range of the iCalendar specification that is required in order to interpret the iCalendar object.
    """

    pass


class CalScale(Property):
    """
    The CALSCALE property defines the calendar scale used for the calendar information specified in the iCalendar
    object.
    """

    pass


class Method(Property):
    """The METHOD property defines the iCalendar object method associated with the calendar object."""

    pass


class Class(Property):
    """The CLASS property defines the access classification for a calendar component."""

    pass


class Description(Property):
    """
    The DESCRIPTION property provides a more complete description of the calendar component than that provided by
    the "SUMMARY" property.
    """

    pass


class Location(Property):
    """The LOCATION property defines the intended venue for the activity defined by a calendar component."""

    pass


class Status(Property):
    """The STATUS property defines the overall status or confirmation for the calendar component."""

    pass


class TimeTransparency(Property):
    """The TRANSP property defines whether an event is transparent to busy time searches."""

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *TIMETRANSPARANCY* but *TRANSP*."""
        return "TRANSP"


class URL(Property):
    """The URL property defines a Uniform Resource Locator (URL) associated with the iCalendar object."""

    pass


class Attach(Property):
    """The ATTACH property provides the capability to associate a document object with a calendar component."""

    pass


class Categories(Property):
    """The CATEGORIES property defines the categories for a calendar component."""

    pass


class Contact(Property):
    """
    The CONTACT property is used to represent contact information or alternately a reference to contact information
    associated with the calendar component.
    """

    pass


class RequestStatus(Property):
    """The REQUEST-STATUS property defines the status code returned for a scheduling request."""

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *REQUESTSTATUS* but *REQUEST-STATUS*."""
        return "REQUEST-STATUS"


class RelatedTo(Property):
    """
    The RELATED-TO property is used to represent a relationship or reference between one calendar component and another.
    """

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *RELATEDTO* but *RELATED-TO*."""
        return "RELATED-TO"


class Resources(Property):
    """
    The RESOURCES property defines the equipment or resources anticipated for an activity specified by a calendar
    component.
    """

    pass


class Action(Property):
    """The ACTION property defines the action to be invoked when an alarm is triggered."""

    pass


class UID(Property):
    """The UID property defines the persistent, globally unique identifier for the calendar component."""

    pass


class Comment(Property):
    """The COMMENT property specifies non-processing information intended to provide a comment to the calendar user."""

    pass


class TZName(Property):
    """The TZNAME property specifies the customary designation for a time zone description."""

    pass


class TZID(Property):
    """
    The TZID property specifies the text value that uniquely identifies the "VTIMEZONE" calendar component in the scope
    of an iCalendar object.
    """

    pass


class TZURL(Property):
    """
    The TZURL property provides a means for a "VTIMEZONE" component to point to a network location that can be used to
    retrieve an up- to-date version of itself.
    """

    pass


class Summary(Property):
    """
    The SUMMARY property defines a short summary or subject for the calendar component.
    """

    pass



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ical_library.base_classes.component import Component


class CalendarParentRelationError(ValueError):
    """Indicate finding the tree root failed as it did not find a VCalendar root."""

    pass


class VEventExpansionFailed(ValueError):
    """Indicate the expansion based on recurring properties failed."""

    pass


class MissingRequiredProperty(ValueError):
    """Indicate a required property is not set for a Component."""

    def __init__(self, component: "Component", missing_property_name: str):
        self.component = component
        self.missing_property_name = missing_property_name

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return (
            f"The required property named {self.missing_property_name} was not set for "
            f"{self.component.__class__.__name__}"
        )



from typing import Iterator, List, Optional, Set, Tuple, Union

from pendulum import Date, DateTime, Duration
from pendulum.tz.timezone import Timezone

from ical_library.help_modules import dt_utils
from ical_library.help_modules.timespan import Timespan
from ical_library.ical_properties.periods import EXDate, RDate
from ical_library.ical_properties.rrule import RRule


def _handle_tz(dt: Union[DateTime, Date], make_tz_aware: Optional[Timezone]) -> Union[DateTime, Date]:
    """
    Small helper function to help with handling timezones.
    It makes the *dt* timezone aware if *make_tz_aware* is specified to be anything but None.
    """
    if make_tz_aware:
        return dt_utils.convert_time_object_to_aware_datetime(dt, make_tz_aware)
    else:
        return dt


def _compute_exdates(
    exdate_list: List[EXDate],
    return_range: Timespan,
    first_event_duration: Duration,
    make_tz_aware: Optional[Timezone],
    starts_to_exclude: Optional[Union[List[Date], List[DateTime]]] = None,
) -> Union[Set[Date], Set[DateTime]]:
    """
    Return a set of Dates or a set of DateTimes that should be excluded based on the EXDate definitions.

    :param exdate_list: The list of EXDate definitions of the component. This defines what dates/datetimes should be
    excluded.
    :param return_range: Returns all DateTimes starts that are within this range.
    :param first_event_duration: The duration of the original component which we should use for the rest of the
    components.
    :param make_tz_aware: If this contains a non-none value, we make all timezone naive DateTimes timezone aware to
    using this Timezone.
    :param starts_to_exclude: List of start Dates or list of start DateTimes of which we already know we should exclude
    them from our recurrence computation (as they have been completely redefined in another element).
    :return: a set of Dates or a set of DateTimes that should be excluded.
    """
    full_list: Union[Set[Date], Set[DateTime]] = set(starts_to_exclude or [])
    for exdate in exdate_list:
        full_list.update(
            _handle_tz(time, make_tz_aware)
            for time in exdate.excluded_date_times
            if return_range.intersects(Timespan(time, time + first_event_duration))  # type: ignore
        )
    return full_list


def _yield_rdate_list(
    rdate_list: List[RDate],
    excluded_times_set: Union[Set[Date], Set[DateTime]],
    first_event_duration: Duration,
    return_range: Timespan,
    make_tz_aware: Optional[Timezone],
) -> Union[Iterator[Tuple[Date, Date]], Iterator[Tuple[DateTime, DateTime]]]:
    """
    Yield a Tuple containing two Dates or yields a tuple containing two DateTimes based on the RDate definitions.

    Note: Instead of the other two functions, this yield a tuple as it might contain a start and end date (or
    start and duration to be more precise) instead of just a start date like the other two. As this duration might be
    different from the original component's duration, we need to always return the start and end date of the component
    here.

    :param rdate_list: The list of RDate definitions of the component. RDate defines a sequence of date/datetimes at
    which the component occurs as well.
    :param excluded_times_set: The set of Dates or a set of DateTimes that should be excluded.
    :param first_event_duration: The duration of the original component which we should use for the rest of the
    components.
    :param return_range: Returns all DateTimes starts that are within this range.
    :param make_tz_aware: If this contains a non-none value, we make all timezone naive DateTimes timezone aware to
    using this Timezone.
    :return: Yields a Tuple containing two Dates or yields a tuple containing two DateTimes representing the full
    duration of the occurrence.
    """
    for rdate in rdate_list:
        rdate_time: Union[DateTime, Date, Tuple[DateTime, DateTime]]
        for rdate_time in rdate.all_values:
            rdate_timespan = Timespan(rdate_time, rdate_time + first_event_duration)
            if not return_range.intersects(rdate_timespan):
                continue
            # Date covers both DateTime and Date. This is because DateTime inherits from Date.
            if isinstance(rdate_time, Date):
                rdate_time = _handle_tz(rdate_time, make_tz_aware=make_tz_aware)
                if return_range.intersects(rdate_timespan) and rdate_time not in excluded_times_set:
                    yield rdate_time, rdate_time + first_event_duration
            elif isinstance(rdate_time, tuple):
                rdate_start = _handle_tz(rdate_time[0], make_tz_aware=make_tz_aware)
                rdate_end = _handle_tz(rdate_time[1], make_tz_aware=make_tz_aware)
                if return_range.intersects(rdate_timespan) and rdate_start not in excluded_times_set:
                    yield rdate_start, rdate_end
            else:
                raise TypeError(f"RDate returned a weird type: {type=}.")


def _yield_rrule_list(
    rrule: Optional[RRule],
    excluded_times_set: Union[Set[Date], Set[DateTime]],
    first_event_start: Union[DateTime, Date],
    first_event_duration: Duration,
    return_range: Timespan,
    make_tz_aware: Optional[Timezone],
) -> Union[Iterator[Date], Iterator[DateTime]]:
    """
    Yield dates or yields DateTimes according to the RRule definition.

    Note: Depending on whether *first_event_start* is a DateTime or a Date, this function also yields in the same
    type.
    :param rrule: The RRule definition of the component.
    :param excluded_times_set: The set of Dates or a set of DateTimes that should be excluded.
    :param first_event_start: The starting point from which we should compute the recurrence.
    :param first_event_duration: The duration of the original component which we should use for the rest of the
    components.
    :param return_range: Returns all DateTimes starts that are within this range.
    :param make_tz_aware: If this contains a non-none value, we make all timezone naive DateTimes timezone aware to
    using this Timezone.
    :return: Yields dates or yields DateTimes for all occurrences of the component specified in the RRule within the
    *return_range*.
    """
    if not rrule:
        return

    iterator = rrule.sequence_iterator(first_event_start, return_range.get_end_in_same_type(first_event_start))
    rrule_time: Union[Date, DateTime]
    for rrule_time in iterator:
        rrule_time = _handle_tz(rrule_time, make_tz_aware=make_tz_aware)
        if not return_range.intersects(Timespan(rrule_time, rrule_time + first_event_duration)):
            continue
        # Date covers both DateTime and Date. This is because DateTime inherits from Date.
        if isinstance(rrule_time, Date):
            if rrule_time not in excluded_times_set:
                yield rrule_time
        else:
            raise TypeError(f"RRule returned a weird type: {type=}")


def expand_event_in_range_only_return_first(
    rdate_list: List[RDate],
    rrule: Optional[RRule],
    first_event_start: Union[DateTime, Date],
    return_range: Timespan,
    make_tz_aware: Optional[Timezone],
) -> Union[Iterator[DateTime], Iterator[Date]]:
    """
    Expand a Component without a duration according to the variables starting from the first_event_start.

    For us to expand events according to the iCalendar specification, we need to keep track of the start times as
    we should not return the same event twice. This is because RDate and RRule might return the same time twice, in
    which case RDate takes priority over RRule.

    :param rdate_list: The list of RDate definitions of the component. RDate defines a sequence of date/datetimes at
    which the component occurs as well
    :param rrule: The RRule definition of the component. RRule specifies a recurring formula from which you can compute
    the sequence of date/datetimes at which the component occurs as well.
    :param first_event_start: The starting point from which we should compute the recurrence.
    :param return_range: Returns all DateTimes starts that are within this range.
    :param make_tz_aware: If this contains a non-none value, we make all timezone naive DateTimes timezone aware to
    using this Timezone.
    :return: an Iterator returning either DateTimes or Dates.
    """
    excluded_times_set: Union[Set[DateTime], Set[Date]] = set()
    iterator = _yield_rdate_list(
        rdate_list=rdate_list,
        excluded_times_set=excluded_times_set,
        first_event_duration=Duration(),
        return_range=return_range,
        make_tz_aware=make_tz_aware,
    )
    rdate: Union[Tuple[Date, Date], Tuple[DateTime, DateTime]]
    for rdate_start, rdate_ending in iterator:
        yield rdate_start

    iterator = _yield_rrule_list(
        rrule=rrule,
        excluded_times_set=set(),
        first_event_start=first_event_start,
        first_event_duration=Duration(),
        return_range=return_range,
        make_tz_aware=make_tz_aware,
    )
    rrule_time: Union[Date, DateTime]
    for rrule_time in iterator:
        yield rrule_time


def expand_component_in_range(
    exdate_list: List[EXDate],
    rdate_list: List[RDate],
    rrule: Optional[RRule],
    first_event_start: Union[DateTime, Date],
    first_event_duration: Duration,
    starts_to_exclude: Union[List[Date], List[DateTime]],
    return_range: Timespan,
    make_tz_aware: Optional[Timezone],
) -> Union[Iterator[Tuple[DateTime, DateTime]], Iterator[Tuple[Date, Date]]]:
    """
    Expand a Component with a duration according to the variables starting from the first_event_start.

    For us to expand events according to the iCalendar specification, we need to keep track of the start times as
    we should not return the same event twice. This is because RDate and RRule might return the same time twice, in
    which case RDate takes priority over RRule. Furthermore, EXDate (which excludes dates from the other sequences)
    takes priority over the other two.

    :param exdate_list: The list of EXDate definitions of the component. This defines what dates/datetimes should be
    excluded.
    :param rdate_list: The list of RDate definitions of the component. RDate defines a sequence of date/datetimes at
    which the component occurs as well.
    :param rrule: The RRule definition of the component. RRule specifies a recurring formula from which you can compute
    the sequence of date/datetimes at which the component occurs as well.
    :param first_event_start: The starting point from which we should compute the recurrence.
    :param first_event_duration: The duration of the original component which we should use for the rest of the
    components.
    :param starts_to_exclude: List of start Dates or list of start DateTimes of which we already know we should exclude
    them from our recurrence computation (as they have been completely redefined in another element).
    :param return_range: Returns all DateTimes starts that are within this range.
    :param make_tz_aware: If this contains a non-none value, we make all timezone naive DateTimes timezone aware to
    using this Timezone.
    :return: an Iterator returning a tuple with two values of either DateTimes or Dates.
    """
    excluded_times_set: Union[Set[DateTime], Set[Date]] = _compute_exdates(
        exdate_list=exdate_list,
        return_range=return_range,
        make_tz_aware=make_tz_aware,
        first_event_duration=first_event_duration,
        starts_to_exclude=starts_to_exclude,
    )

    iterator = _yield_rdate_list(
        rdate_list=rdate_list,
        excluded_times_set=excluded_times_set,
        first_event_duration=first_event_duration,
        return_range=return_range,
        make_tz_aware=make_tz_aware,
    )
    rdate: Union[Tuple[Date, Date], Tuple[DateTime, DateTime]]
    for rdate in iterator:
        yield rdate

    iterator = _yield_rrule_list(
        rrule=rrule,
        excluded_times_set=excluded_times_set,
        first_event_start=first_event_start,
        first_event_duration=first_event_duration,
        return_range=return_range,
        make_tz_aware=make_tz_aware,
    )
    rrule_time: Union[Date, DateTime]
    for rrule_time in iterator:
        yield rrule_time, rrule_time + first_event_duration



from typing import Optional

from ical_library.base_classes.component import Component
from ical_library.exceptions import MissingRequiredProperty
from ical_library.ical_properties.ical_duration import ICALDuration
from ical_library.ical_properties.ints import Repeat
from ical_library.ical_properties.pass_properties import Action, Attach
from ical_library.ical_properties.trigger import Trigger


class VAlarm(Component):
    """
    This class represents the VAlarm component specified in RFC 5545 in '3.6.6. Alarm Component'.

    A "VALARM" calendar component is a grouping of component properties that is a reminder or alarm for an event or a
    to-do. For example, it may be used to define a reminder for a pending event or an overdue to-do.
    The "VALARM" calendar component MUST only appear within either a "VEVENT" or "VTODO" calendar component

    :param action: The Action property. Required and must occur exactly once.
    :param trigger: The Trigger property. Required and must occur exactly once.
    :param duration: The ICALDuration property. Optional, but may occur at most once. If this item is
        present, repeat may not be present.
    :param repeat: The Repeat property. Optional, but may occur at most once. If this item is
        present, duration may not be present.
    :param attach: The Attach property. Optional, but may occur at most once.
    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    """

    def __init__(
        self,
        action: Optional[Action] = None,
        trigger: Optional[Trigger] = None,
        duration: Optional[ICALDuration] = None,
        repeat: Optional[Repeat] = None,
        attach: Optional[Attach] = None,
        parent: Optional[Component] = None,
    ):
        super().__init__("VALARM", parent=parent)

        # Required
        self._action: Optional[Action] = self.as_parent(action)
        self._trigger: Optional[Trigger] = self.as_parent(trigger)

        # Both optional and may only occur once. But if one occurs, the other also has to occur.
        self.duration: Optional[ICALDuration] = self.as_parent(duration)
        self.repeat: Optional[Repeat] = self.as_parent(repeat)

        # Optional, may only occur once
        self.attach: Optional[Attach] = self.as_parent(attach)

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"VAlarm({self.action.value}: {self.trigger.value})"

    @property
    def action(self) -> Action:
        """A getter to ensure the required property is set."""
        if self._action is None:
            raise MissingRequiredProperty(self, "action")
        return self._action

    @action.setter
    def action(self, value: Action):
        """A setter to set the required property."""
        self._action = value

    @property
    def trigger(self) -> Trigger:
        """Getter that ensures the required property is set."""
        if self._trigger is None:
            raise MissingRequiredProperty(self, "trigger")
        return self._trigger

    @trigger.setter
    def trigger(self, value: Trigger):
        """A setter to set the required property."""
        self._trigger = value



import functools
import weakref


def instance_lru_cache(*lru_args, **lru_kwargs):
    """
    Add a lru_cache on an instance function without causing a memory leak.
    At the moment you add a standard lru_cache on an instance function (e.g. `def abc(self):`) it counts a reference for
     the python reference counter. This means that when the instance is deleted in the users code, the instance will not
     be collected by the garbage collector because the lru_cache keeps the reference counter for the instance at a
     positive value.
    This implementation prevents this by ensuring the reference to the self attribute is a weak reference. Thereby it is
     not counted in the reference counter and thus can be freely cleared by the garbage collector.
    Inspiration from: https://stackoverflow.com/a/33672499/2277445
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapped_func(self, *args, **kwargs):
            # We're storing the wrapped method inside the instance. If we had
            # a strong reference to self the instance would never die.
            self_weak = weakref.ref(self)

            @functools.wraps(func)
            @functools.lru_cache(*lru_args, **lru_kwargs)
            def cached_method(*args, **kwargs):
                return func(self_weak(), *args, **kwargs)

            return cached_method(*args, **kwargs)

        return wrapped_func

    return decorator



from typing import Literal, Union

import pendulum
from pendulum import DateTime, Duration

from ical_library.base_classes.property import Property


class Trigger(Property):
    """The TRIGGER property specifies when an alarm will trigger."""

    @property
    def kind(self) -> Literal["DATE-TIME", "DURATION"]:
        """Return the type of the property value."""
        kind_of_value = self.get_property_parameter("VALUE")
        return "DATE-TIME" if kind_of_value and kind_of_value == "DATE-TIME" else "DURATION"  # noqa

    def parse_value(self) -> Union[Duration, DateTime]:
        """Parse the value of this property based on the VALUE property parameter."""
        if self.kind == "DURATION":
            parsed_value: Duration = pendulum.parse(self.value)
            if not isinstance(parsed_value, Duration):
                raise TypeError(f"Invalid value passed for Duration: {self.value=}")
            return parsed_value
        else:
            parsed_value: DateTime = pendulum.parse(self.value)
            if not isinstance(parsed_value, DateTime):
                raise TypeError(f"Invalid value passed for DateTime: {self.value=}")
            return parsed_value

    def trigger_relation(self) -> Literal["START", "END"]:
        """Get the trigger relation, whether the duration should be relative to the start or the end of a component."""
        return "START" if self.get_property_parameter_default("RELATED", "START") == "START" else "END"  # noqa



from typing import List

from pendulum import DateTime
from pendulum.tz.zoneinfo.transition import Transition

from ical_library.ical_components import VCalendar, VTimeZone


def filter_transitions(start: DateTime, end: DateTime, transitions: List[Transition]) -> List[Transition]:
    start_epoch = int(start.timestamp())
    end_epoch = int(end.timestamp())
    return [transition for transition in transitions if start_epoch <= transition.at <= end_epoch]


def test_get_ordered_timezone_overview_as_transition(berlin_timezone_calendar: VCalendar):
    """Test that the timezone components are correctly parsed."""
    # For reference material, you can use the transitions that are present in this;
    # berlin = pendulum._safe_timezone("Europe/Berlin")
    berlin_timezone: VTimeZone = berlin_timezone_calendar.get_timezone("Europe/Berlin")
    generated_transitions = berlin_timezone.get_ordered_timezone_overview_as_transition(DateTime(2023, 1, 1))
    year_1970 = filter_transitions(DateTime(1970, 1, 1), DateTime(1971, 1, 1), generated_transitions)
    march_1970_switch = year_1970[0]
    october_1970_switch = year_1970[1]
    assert march_1970_switch.at == 7520400
    assert march_1970_switch.ttype.offset == 7200
    assert march_1970_switch.ttype.is_dst() is True
    assert march_1970_switch.ttype.abbreviation == "CEST"
    assert march_1970_switch.previous is None
    assert october_1970_switch.at == 25664400
    assert october_1970_switch.ttype.offset == 3600
    assert october_1970_switch.ttype.is_dst() is False
    assert october_1970_switch.ttype.abbreviation == "CET"
    assert october_1970_switch.previous.at == 7520400
    assert october_1970_switch.previous.ttype.offset == 7200
    assert october_1970_switch.previous.ttype.is_dst() is True
    assert october_1970_switch.previous.ttype.abbreviation == "CEST"
    assert october_1970_switch.previous.previous is None

    year_2022 = filter_transitions(DateTime(2022, 1, 1), DateTime(2023, 1, 1), generated_transitions)
    march_2022_switch = year_2022[0]
    october_2022_switch = year_2022[1]

    assert march_2022_switch.at == 1648342800
    assert march_2022_switch.ttype.offset == 7200
    assert march_2022_switch.ttype.is_dst() is True
    assert march_2022_switch.ttype.abbreviation == "CEST"

    assert october_2022_switch.at == 1667091600
    assert october_2022_switch.ttype.offset == 3600
    assert october_2022_switch.ttype.is_dst() is False
    assert october_2022_switch.ttype.abbreviation == "CET"


def test_custom_timezone(berlin_timezone_calendar: VCalendar):
    """Test that the offset switches correctly when there is a time change (which is around 2022-03-27)."""
    tz = berlin_timezone_calendar.get_timezone("Europe/Berlin").get_as_timezone_object(DateTime(2023, 1, 1))
    assert DateTime(2022, 1, 1).in_timezone(tz).offset == 3600
    assert DateTime(2022, 4, 1).in_timezone(tz).offset == 7200



from typing import List, Optional

from ical_library.base_classes.component import Component
from ical_library.exceptions import MissingRequiredProperty
from ical_library.help_modules.timespan import TimespanWithParent
from ical_library.ical_properties.cal_address import Attendee, Organizer
from ical_library.ical_properties.dt import DTEnd, DTStamp, DTStart
from ical_library.ical_properties.pass_properties import Comment, Contact, RequestStatus, UID, URL
from ical_library.ical_properties.periods import FreeBusyProperty


class VFreeBusy(Component):
    """
    This class represents the VFREEBUSY component specified in RFC 5545 in '3.6.4. Free/Busy Component'.

    A "VFREEBUSY" calendar component is a grouping of component properties that represents either a request for free
    or busy time information, a reply to a request for free or busy time information, or a published set of busy time
    information.

    :param name: The actual name of this component instance. E.g. VEVENT, RRULE, VCUSTOMCOMPONENT.
    :param dtstamp: The DTStamp property. Required and must occur exactly once.
    :param uid: The UID property. Required and must occur exactly once.
    :param contact: The Contact property. Optional, but may occur at most once.
    :param dtstart: The DTStart property. Optional, but may occur at most once.
    :param dtend: The DTEnd property. Optional, but may occur at most once.
    :param organizer: The Organizer property. Optional, but may occur at most once.
    :param url: The URL property. Optional, but may occur at most once.
    :param attendee: The Attendee property. Optional, but may occur multiple times.
    :param comment: The Comment property. Optional, but may occur multiple times.
    :param freebusy: The FreeBusyProperty property. Optional, but may occur multiple times.
    :param rstatus: The RequestStatus property. Optional, but may occur multiple times.
    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    """

    def __init__(
        self,
        dtstamp: Optional[DTStamp] = None,
        uid: Optional[UID] = None,
        contact: Optional[Contact] = None,
        dtstart: Optional[DTStart] = None,
        dtend: Optional[DTEnd] = None,
        organizer: Optional[Organizer] = None,
        url: Optional[URL] = None,
        attendee: Optional[List[Attendee]] = None,
        comment: Optional[List[Comment]] = None,
        freebusy: Optional[List[FreeBusyProperty]] = None,
        rstatus: Optional[List[RequestStatus]] = None,
        parent: Optional[Component] = None,
    ):
        super().__init__("VFREEBUSY", parent=parent)

        # Required
        self._dtstamp: Optional[DTStamp] = self.as_parent(dtstamp)
        self._uid: Optional[UID] = self.as_parent(uid)

        # Optional, may only occur once
        self.contact: Optional[Contact] = self.as_parent(contact)
        self.dtstart: Optional[DTStart] = self.as_parent(dtstart)
        self.dtend: Optional[DTEnd] = self.as_parent(dtend)
        self.organizer: Optional[Organizer] = self.as_parent(organizer)
        self.url: Optional[URL] = self.as_parent(url)

        # Optional, may occur more than once
        self.attendee: Optional[List[Attendee]] = self.as_parent(attendee)
        self.comment: Optional[List[Comment]] = self.as_parent(comment)
        self.freebusy: Optional[List[FreeBusyProperty]] = self.as_parent(freebusy)
        self.rstatus: Optional[List[RequestStatus]] = self.as_parent(rstatus)

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"VFreeBusy({self.dtstart.value if self.dtstart else ''}, {self.dtend.value if self.dtend else ''})"

    @property
    def dtstamp(self) -> DTStamp:
        """A getter to ensure the required property is set."""
        if self._dtstamp is None:
            raise MissingRequiredProperty(self, "dtstamp")
        return self._dtstamp

    @dtstamp.setter
    def dtstamp(self, value: DTStamp):
        """A setter to set the required property."""
        self._dtstamp = value

    @property
    def uid(self) -> UID:
        """A getter to ensure the required property is set."""
        if self._uid is None:
            raise MissingRequiredProperty(self, "uid")
        return self._uid

    @uid.setter
    def uid(self, value: UID):
        """A setter to set the required property."""
        self._uid = value

    @property
    def timespan(self) -> Optional[TimespanWithParent]:
        start = self.dtstart.datetime_or_date_value if self.dtstart else None
        end = self.dtend.datetime_or_date_value if self.dtend else None
        if start is None:
            return None
        if end is None:
            TimespanWithParent(parent=self, begin=start, end=start)
        return TimespanWithParent(parent=self, begin=start, end=end)



from ical_library.base_classes.property import Property


class _IntProperty(Property):
    """This property class should be inherited. It represents a property that contain just an int as value."""

    @property
    def int_value(self) -> int:
        """Return the value as an int."""
        return int(self.value)


class Priority(_IntProperty):
    """The PRIORITY property represents the relative priority for a calendar component."""

    pass


class Sequence(_IntProperty):
    """
    The SEQUENCE property defines the revision sequence number of the calendar component within a sequence of revisions.
    """

    pass


class Repeat(_IntProperty):
    """The REPEAT property defines the number of times the alarm should be repeated, after the initial trigger."""

    pass


class PercentComplete(_IntProperty):
    """
    The PERCENT-COMPLETE property is used by an assignee or delegatee of a to-do to convey the percent completion of
    a to-do to the "Organizer".
    """

    @property
    def percentage(self) -> int:
        return self.int_value

    @classmethod
    def get_ical_name_of_class(cls) -> str:
        """Overwrite the iCal name of this class as it is not *PERCENTCOMPLETE* but *PERCENT-COMPLETE*."""
        return "PERCENT-COMPLETE"



from typing import List, Optional, TYPE_CHECKING

from pendulum import DateTime

from ical_library.base_classes.component import Component
from ical_library.exceptions import MissingRequiredProperty
from ical_library.ical_components.v_event import VEvent
from ical_library.ical_components.v_free_busy import VFreeBusy
from ical_library.ical_components.v_journal import VJournal
from ical_library.ical_components.v_timezone import VTimeZone
from ical_library.ical_components.v_todo import VToDo
from ical_library.ical_properties.pass_properties import CalScale, Method, ProdID, Version

if TYPE_CHECKING:
    from ical_library.timeline import Timeline


class VCalendar(Component):
    """
    This class represents the VCALENDAR component specified in RFC 5545 in '3.6. Calendar Components'.

    The "VCALENDAR" component consists of a sequence of calendar properties and one or more calendar components.
    The calendar properties are attributes that apply to the calendar object as a whole. The calendar components are
    collections of properties that express a particular calendar semantic. For example, the calendar component can
    specify an event, a to-do, a journal entry, time zone information, free/busy time information, or an alarm.

    :param prodid: The ProdID property. Required and must occur exactly once.
    :param version: The Version property. Required and must occur exactly once.
    :param calscale: The CalScale property. Optional, but may occur at most once.
    :param method: The Method property. Optional, but may occur at most once.
    :param events: Optional list of VEvent components. Each component may occur multiple times.
    :param todos: Optional list of VToDo components. Each component may occur multiple times.
    :param journals: Optional list of VJournal components. Each component may occur multiple times.
    :param free_busy_list: Optional list of VFreeBusy components. Each component may occur multiple times.
    :param time_zones: Optional list of VTimeZone components. Each component may occur multiple times.
    """

    def __init__(
        self,
        prodid: Optional[ProdID] = None,
        version: Optional[Version] = None,
        calscale: Optional[CalScale] = None,
        method: Optional[Method] = None,
        events: Optional[List[VEvent]] = None,
        todos: Optional[List[VToDo]] = None,
        journals: Optional[List[VJournal]] = None,
        free_busy_list: Optional[List[VFreeBusy]] = None,
        time_zones: Optional[List[VTimeZone]] = None,
    ):
        super().__init__("VCALENDAR")

        # Required properties, only one occurrence allowed.
        self._prodid: Optional[ProdID] = self.as_parent(prodid)
        self._version: Optional[Version] = self.as_parent(version)

        # Optional properties, must not occur more than once.
        self.calscale: Optional[CalScale] = self.as_parent(calscale)
        self.method: Optional[Method] = self.as_parent(method)

        # These are children Components
        self.events: List[VEvent] = events or []
        self.todos: List[VToDo] = todos or []
        self.journals: List[VJournal] = journals or []
        self.free_busy_list: List[VFreeBusy] = free_busy_list or []
        self.time_zones: List[VTimeZone] = time_zones or []

        # Only the VCalender stores the entire list.
        self._lines: Optional[List[str]] = None

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"VCalendar({self.prodid.value}, {self.version.value})"

    @property
    def prodid(self) -> ProdID:
        """A getter to ensure the required property is set."""
        if self._prodid is None:
            raise MissingRequiredProperty(self, "prodid")
        return self._prodid

    @prodid.setter
    def prodid(self, value: ProdID):
        """A setter to set the required property."""
        self._prodid = value

    @property
    def version(self) -> Version:
        """A getter to ensure the required property is set."""
        if self._version is None:
            raise MissingRequiredProperty(self, "version")
        return self._version

    @version.setter
    def version(self, value: Version):
        """A setter to set the required property."""
        self._version = value

    @property
    def calendar_scale(self) -> str:
        """Return the calendar scale according to RFC 5545."""
        return self.calscale.value if self.calscale else "GREGORIAN"

    def get_timezone(self, tzid: str) -> VTimeZone:
        """Get the corresponding VTimeZone object based on the given timezone identifier."""
        for timezone in self.time_zones:
            if timezone.tzid.value == tzid:
                return timezone
        raise ValueError(f"Could not find Timezone with {tzid=}.")

    def get_aware_dt_for_timezone(self, dt: DateTime, tzid: str) -> DateTime:
        """Return the timezone aware DateTime object for a given TimeZone identifier."""
        return self.get_timezone(tzid).convert_naive_datetime_to_aware(dt)

    @property
    def timeline(self) -> "Timeline":
        """Return a timeline of VEvents from 1970-00-00T00:00:00 to 2100-00-00T00:00:00."""
        from ical_library.timeline import Timeline

        return Timeline(self)

    def get_limited_timeline(self, start: Optional[DateTime], end: Optional[DateTime]) -> "Timeline":
        """
        Return a timeline of VEvents limited by *start* and *end*

        :param start: Only include events in the timeline with a starting date later than this value.
        :param end: Only include events in the timeline with a starting date earlier than this value.
        """
        from ical_library.timeline import Timeline

        return Timeline(self, start, end)

    def parse_component(self, lines: List[str], line_number: int) -> int:
        """
        Parse a new component in the RAW string list.
        :param lines: A list of all the lines in the iCalendar file.
        :param line_number: The line number at which this component starts.
        :return: The line number at which this component ends.
        """
        self._lines = lines
        return super().parse_component(lines=lines, line_number=line_number)

    def get_original_ical_text(self, start_line: int, end_line: int) -> str:
        """
        Get the original iCAL text for your property from the RAW string list.
        :param start_line: The starting line index for the component you wish to show.
        :param end_line: The ending line index for the component you wish to show.
        :return: The complete string, as it was in the RAW string list, for the component you wish to show.
        """
        lines = self._lines
        if lines is None:
            raise TypeError("We should first parse the component section before calling this function.")
        return "\n".join(line for line in self.tree_root._lines[max(0, start_line) : min(len(lines), end_line)])



from typing import List

import pendulum
from pendulum import DateTime

from ical_library.help_modules.timespan import Timespan, TimespanWithParent
from ical_library.ical_components import VCalendar


def test_recurrence_with_offset_changes(multi_offset_calendar: VCalendar) -> None:
    """
    This test verifies that calendars that have a recurrence set, correctly handle time zone changes.
    In this case we have Amsterdam time zone which changes from +02:00 to +01:00 at 2022-10-30
    """
    recurring_event = multi_offset_calendar.events[0]
    return_rage = Timespan(DateTime(2022, 10, 1), DateTime(2022, 11, 10))
    components: List[TimespanWithParent] = list(recurring_event.expand_component_in_range(return_rage, []))
    assert components[0].begin == pendulum.parse("2022-10-18T12:00:00+02:00")
    assert components[0].end == pendulum.parse("2022-10-18T17:00:00+02:00")
    assert components[1].begin == pendulum.parse("2022-10-25T12:00:00+02:00")
    assert components[1].end == pendulum.parse("2022-10-25T17:00:00+02:00")
    assert components[2].begin == pendulum.parse("2022-11-01T12:00:00+01:00")
    assert components[2].end == pendulum.parse("2022-11-01T17:00:00+01:00")
    assert components[3].begin == pendulum.parse("2022-11-08T12:00:00+01:00")
    assert components[3].end == pendulum.parse("2022-11-08T17:00:00+01:00")



import functools
from typing import Any, Optional, Tuple, TYPE_CHECKING, Union

from pendulum import Date, DateTime

from ical_library.help_modules import dt_utils

if TYPE_CHECKING:
    from ical_library.base_classes.component import Component


@functools.total_ordering
class Timespan:
    """
    Represents a time span with a beginning and end date.

    Contains multiple handy functions when working with a time span like the is_included_in or intersect function.
    :param begin: The beginning of the timespan.
    :param end: The end of the timespan.
    """

    def __init__(self, begin: Union[Date, DateTime], end: Union[Date, DateTime]):
        """For simplicity, we convert the variables to timezone aware :class:`pendulum.DateTime` instances."""
        self.begin: DateTime = dt_utils.convert_time_object_to_aware_datetime(begin)
        self.end: DateTime = dt_utils.convert_time_object_to_aware_datetime(end)

    def __repr__(self) -> str:
        return f"Timespan({self.begin}, {self.end})"

    @property
    def tuple(self) -> Tuple[DateTime, DateTime]:
        """
        Return the timespan as a tuple.
        :return: The beginning and end in a tuple format.
        """
        return self.begin, self.end

    def get_end_in_same_type(self, time: Union[Date, DateTime]) -> Union[Date, DateTime]:
        """
        Return *self.end* in the same format as *time*.
        :param time: A time format in which we would like to return the *self.end*.
        :return: Return *self.end* in the same format as *time*.
        """
        if not isinstance(time, DateTime):
            return Date(self.end.year, self.end.month, self.end.day)
        elif time.tz is None:  # At this point we know it is a DateTime object.
            return self.end.replace(tzinfo=None)  # type: ignore
        return self.end.in_timezone(time.tz)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Timespan):
            return self.tuple == other.tuple
        else:
            return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Timespan):
            return self.tuple < other.tuple
        else:
            return NotImplemented

    def __gt__(self, other: Any) -> bool:
        if isinstance(other, Timespan):
            return self.tuple > other.tuple
        else:
            return NotImplemented

    def __le__(self, other: Any) -> bool:
        if isinstance(other, Timespan):
            return self.tuple <= other.tuple
        else:
            return NotImplemented

    def __ge__(self, other: Any) -> bool:
        if isinstance(other, Timespan):
            return self.tuple >= other.tuple
        else:
            return NotImplemented

    def is_included_in(self, other: "Timespan") -> bool:
        """
        Check whether the *other* timespan is included within this timespan.
        :param other: Another timespan.
        :return: Boolean whether the *other* timespan is included within this timespan.
        """
        return other.begin <= self.begin and self.end <= other.end

    def intersects(self, other: "Timespan") -> bool:
        """
        Check whether the *other* timespan intersects with this timespan.
        :param other: Another timespan.
        :return: boolean whether the *other* timespan intersects with this timespan.
        """
        return (
            self.begin <= other.begin < self.end
            or self.begin <= other.end < self.end
            or other.begin <= self.begin < other.end
            or other.begin <= self.end < other.end
        )

    def includes(self, instant: Union[Date, DateTime]) -> bool:
        """
        Check if the *instant* is included in this timespan.
        :param instant: An instance that represents a moment in time.
        :return: Boolean of whether the *instant* is included in this timespan.
        """
        dt = dt_utils.convert_time_object_to_aware_datetime(instant)
        return self.begin <= dt < self.end


class TimespanWithParent(Timespan):
    """
    Init a Timespan object which represents the length of a specific component(marked as parent).
    :param parent: The component instance which this timespan represents.
    :param begin: The beginning of the timespan.
    :param end: The end of the timespan.
    """

    def __init__(self, parent: Optional["Component"], begin: Union[Date, DateTime], end: Union[Date, DateTime]):
        super().__init__(begin, end)
        self.parent = parent



from ical_library.ical_components import (
    DayLight,
    Standard,
    VAlarm,
    VCalendar,
    VEvent,
    VFreeBusy,
    VJournal,
    VTimeZone,
    VToDo,
)
from ical_library.ical_properties.cal_address import Attendee, Organizer
from ical_library.ical_properties.dt import Completed, Created, DTEnd, DTStamp, DTStart, Due, LastModified, RecurrenceID
from ical_library.ical_properties.geo import GEO
from ical_library.ical_properties.ical_duration import ICALDuration
from ical_library.ical_properties.ints import PercentComplete, Priority, Repeat, Sequence
from ical_library.ical_properties.pass_properties import (
    Action,
    Attach,
    CalScale,
    Categories,
    Class,
    Comment,
    Contact,
    Description,
    Location,
    Method,
    ProdID,
    RelatedTo,
    RequestStatus,
    Resources,
    Status,
    Summary,
    TimeTransparency,
    TZID,
    TZName,
    TZURL,
    UID,
    URL,
    Version,
)
from ical_library.ical_properties.periods import EXDate, FreeBusyProperty, RDate
from ical_library.ical_properties.rrule import RRule
from ical_library.ical_properties.trigger import Trigger
from ical_library.ical_properties.tz_offset import TZOffsetFrom, TZOffsetTo


def test_get_property_ical_names():
    assert VCalendar.get_property_ical_names() == {"calscale", "prodid", "version", "method"}
    assert VAlarm.get_property_ical_names() == {"attach", "repeat", "trigger", "action", "duration"}


def test_valarm_get_property_mapping():
    assert VAlarm._get_property_mapping() == {
        "ACTION": ("action", Action, False),
        "ATTACH": ("attach", Attach, False),
        "DURATION": ("duration", ICALDuration, False),
        "REPEAT": ("repeat", Repeat, False),
        "TRIGGER": ("trigger", Trigger, False),
    }


def test_vcalendar_get_property_mapping():
    assert VCalendar._get_property_mapping() == {
        "CALSCALE": ("calscale", CalScale, False),
        "METHOD": ("method", Method, False),
        "PRODID": ("prodid", ProdID, False),
        "VERSION": ("version", Version, False),
    }


def test_vevent_get_property_mapping():
    assert VEvent._get_property_mapping() == {
        "ATTACH": ("attach", Attach, True),
        "ATTENDEE": ("attendee", Attendee, True),
        "CATEGORIES": ("categories", Categories, True),
        "CLASS": ("ical_class", Class, False),
        "COMMENT": ("comment", Comment, True),
        "CONTACT": ("contact", Contact, True),
        "CREATED": ("created", Created, False),
        "DESCRIPTION": ("description", Description, False),
        "DTEND": ("dtend", DTEnd, False),
        "DTSTAMP": ("dtstamp", DTStamp, False),
        "DTSTART": ("dtstart", DTStart, False),
        "DURATION": ("duration", ICALDuration, False),
        "EXDATE": ("exdate", EXDate, True),
        "GEO": ("geo", GEO, False),
        "LAST-MODIFIED": ("last_modified", LastModified, False),
        "LOCATION": ("location", Location, False),
        "ORGANIZER": ("organizer", Organizer, False),
        "PRIORITY": ("priority", Priority, False),
        "RDATE": ("rdate", RDate, True),
        "RECURRENCE-ID": ("recurrence_id", RecurrenceID, False),
        "RELATED-TO": ("related", RelatedTo, True),
        "REQUEST-STATUS": ("rstatus", RequestStatus, True),
        "RESOURCES": ("resources", Resources, True),
        "RRULE": ("rrule", RRule, False),
        "SEQUENCE": ("sequence", Sequence, False),
        "STATUS": ("status", Status, False),
        "SUMMARY": ("summary", Summary, False),
        "TRANSP": ("transp", TimeTransparency, False),
        "UID": ("uid", UID, False),
        "URL": ("url", URL, False),
    }


def test_vfreebusy_get_property_mapping():
    assert VFreeBusy._get_property_mapping() == {
        "ATTENDEE": ("attendee", Attendee, True),
        "COMMENT": ("comment", Comment, True),
        "CONTACT": ("contact", Contact, False),
        "DTEND": ("dtend", DTEnd, False),
        "DTSTAMP": ("dtstamp", DTStamp, False),
        "DTSTART": ("dtstart", DTStart, False),
        "FREEBUSY": ("freebusy", FreeBusyProperty, True),
        "ORGANIZER": ("organizer", Organizer, False),
        "REQUEST-STATUS": ("rstatus", RequestStatus, True),
        "UID": ("uid", UID, False),
        "URL": ("url", URL, False),
    }


def test_vjournal_get_property_mapping():
    assert VJournal._get_property_mapping() == {
        "ATTACH": ("attach", Attach, True),
        "ATTENDEE": ("attendee", Attendee, True),
        "CATEGORIES": ("categories", Categories, True),
        "CLASS": ("ical_class", Class, False),
        "COMMENT": ("comment", Comment, True),
        "CONTACT": ("contact", Contact, True),
        "CREATED": ("created", Created, False),
        "DESCRIPTION": ("description", Description, True),
        "DTSTAMP": ("dtstamp", DTStamp, False),
        "DTSTART": ("dtstart", DTStart, False),
        "EXDATE": ("exdate", EXDate, True),
        "LAST-MODIFIED": ("last_modified", LastModified, False),
        "ORGANIZER": ("organizer", Organizer, False),
        "RDATE": ("rdate", RDate, True),
        "RECURRENCE-ID": ("recurrence_id", RecurrenceID, False),
        "RELATED-TO": ("related", RelatedTo, True),
        "REQUEST-STATUS": ("rstatus", RequestStatus, True),
        "RRULE": ("rrule", RRule, False),
        "SEQUENCE": ("sequence", Sequence, False),
        "STATUS": ("status", Status, False),
        "SUMMARY": ("summary", Summary, False),
        "UID": ("uid", UID, False),
        "URL": ("url", URL, False),
    }


def test_vtimezone_get_property_mapping():
    assert VTimeZone._get_property_mapping() == {
        "LAST-MODIFIED": ("last_mod", LastModified, False),
        "TZID": ("tzid", TZID, False),
        "TZURL": ("tzurl", TZURL, False),
    }


def test_vtodo_get_property_mapping():
    assert VToDo._get_property_mapping() == {
        "ATTACH": ("attach", Attach, True),
        "ATTENDEE": ("attendee", Attendee, True),
        "CATEGORIES": ("categories", Categories, True),
        "CLASS": ("ical_class", Class, False),
        "COMMENT": ("comment", Comment, True),
        "COMPLETED": ("completed", Completed, False),
        "CONTACT": ("contact", Contact, True),
        "CREATED": ("created", Created, False),
        "DESCRIPTION": ("description", Description, False),
        "DTSTAMP": ("dtstamp", DTStamp, False),
        "DTSTART": ("dtstart", DTStart, False),
        "DUE": ("due", Due, False),
        "DURATION": ("duration", ICALDuration, False),
        "EXDATE": ("exdate", EXDate, True),
        "GEO": ("geo", GEO, False),
        "LAST-MODIFIED": ("last_modified", LastModified, False),
        "LOCATION": ("location", Location, False),
        "ORGANIZER": ("organizer", Organizer, False),
        "PERCENT-COMPLETE": ("percent", PercentComplete, False),
        "PRIORITY": ("priority", Priority, False),
        "RDATE": ("rdate", RDate, True),
        "RECURRENCE-ID": ("recurrence_id", RecurrenceID, False),
        "RELATED-TO": ("related", RelatedTo, True),
        "REQUEST-STATUS": ("rstatus", RequestStatus, True),
        "RESOURCES": ("resources", Resources, True),
        "RRULE": ("rrule", RRule, False),
        "SEQUENCE": ("sequence", Sequence, False),
        "STATUS": ("status", Status, False),
        "SUMMARY": ("summary", Summary, False),
        "UID": ("uid", UID, False),
        "URL": ("url", URL, False),
    }


def test_daylight_and_get_property_mapping():
    expected = {
        "COMMENT": ("comment", Comment, True),
        "DTSTART": ("dtstart", DTStart, False),
        "RDATE": ("rdate", RDate, True),
        "RRULE": ("rrule", RRule, False),
        "TZNAME": ("tzname", TZName, True),
        "TZOFFSETFROM": ("tzoffsetfrom", TZOffsetFrom, False),
        "TZOFFSETTO": ("tzoffsetto", TZOffsetTo, False),
    }
    assert DayLight._get_property_mapping() == expected
    assert Standard._get_property_mapping() == expected



from typing import Dict, Optional, TYPE_CHECKING

from ical_library.base_classes.base_class import ICalBaseClass
from ical_library.help_modules.component_context import ComponentContext
from ical_library.help_modules.lru_cache import instance_lru_cache

if TYPE_CHECKING:
    from ical_library.base_classes.component import Component


class Property(ICalBaseClass):
    """
    This is the base class for any property (according to the RFC 5545 specification) in iCal-library.

    A property always exists of three parts:

    - The name of the property.
    - The property parameters, this is optional and does not need to be present.
    - The value of the property.

    A line containing a property typically has the following format:
    `PROPERTY-NAME;parameterKey=parameterValue,anotherParameterKey=anotherValue:actual-value`

    Any property that is predefined according to the RFC 5545 should inherit this class, e.g. UID, RRule.
    Only x-properties or iana-properties should instantiate the Property class directly.

    :param value: The value of the property.
    :param name: The properties name, e.g. `RRULE`.
    :param property_parameters: The property parameters for this definition.
    :param parent: Instance of the :class:`Component` it is a part of.
    """

    def __init__(
        self,
        value: Optional[str],
        name: Optional[str] = None,
        property_parameters: Optional[str] = None,
        parent: "Component" = None,
    ):
        name = name if self.__class__ == Property else self.__class__.get_ical_name_of_class()
        super().__init__(name=name, parent=parent or ComponentContext.get_current_component())
        if parent is None and self.parent is not None:
            self.parent.set_property(self)

        self._property_parameters: Optional[str] = property_parameters
        self._value: Optional[str] = value

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"{self.__class__.__name__}({self.as_original_string})"

    def __eq__(self: "Property", other: "Property") -> bool:
        """Return whether the current instance and the other instance are the same."""
        if type(self) != type(other):
            return False
        return self.as_original_string == other.as_original_string

    @property
    @instance_lru_cache()
    def property_parameters(self) -> Dict[str, str]:
        """
        Return (and cache) all the property's parameters as a dictionary of strings.

        Note: When the instance is collected by the garbage collection, the cache is automatically deleted as well.

        :return: all the property's parameters as a dictionary of strings
        """
        property_parameters_str = self._property_parameters or ""
        return {
            key_and_value.split("=")[0]: key_and_value.split("=")[1]
            for key_and_value in property_parameters_str.split(",")
            if key_and_value.count("=") == 1
        }

    def has_property_parameter(self, key: str) -> Optional[bool]:
        """
        Return whether this property has a property parameter with a specific *key*.

        :param key: What key to search for.
        :return: boolean whether this property has a property parameter with a specific *key*.
        """
        return key in self.property_parameters

    def get_property_parameter(self, key: str) -> Optional[str]:
        """
        Get a property parameter's value with a specific key.

        :param key: The identifier of the property parameter.
        :return: The requested property parameter, or if that is not present, the default value.
        """
        return self.property_parameters.get(key, None)

    def get_property_parameter_default(self, key: str, default: str) -> str:
        """
        Get a property parameter's value with a specific key, where the default may not be None.

        :param key: The identifier of the property parameter.
        :param default: A value to return when the property parameter is not present, which may not be None.
        :return: The requested property parameter, or if that is not present, the default value.
        """
        return self.property_parameters.get(key, default)

    @property
    def value(self) -> Optional[str]:
        """Return the value of this property."""
        return self._value

    @property
    def as_original_string(self) -> str:
        """
        Return the iCalendar representation of the parameter.
        :return: the iCalendar string representation.
        """
        add_subs = f";{self._property_parameters}" if self._property_parameters else ""
        return f"{self._name}{add_subs}:{self._value}"



from typing import Optional

from ical_library.base_classes.property import Property


class _CalAddress(Property):
    @property
    def persons_name(self) -> Optional[str]:
        """Return the persons name, identified by the CN property parameter."""
        return self.get_property_parameter("CN")

    @property
    def email(self) -> Optional[str]:
        """Return the email if the value starts with `mailto:`. Otherwise return None."""
        if self.value.startswith("mailto:"):
            return self.value[len("mailto:") :]
        return None

    @property
    def cu_type(self) -> str:
        """Return the CUTYPE."""
        return self.get_property_parameter_default("CUTYPE", default="INDIVIDUAL")

    @property
    def member(self) -> Optional[str]:
        """Return the membership property parameter."""
        return self.get_property_parameter("MEMBER")

    @property
    def role(self) -> str:
        """Return the role of the person."""
        return self.get_property_parameter_default("ROLE", default="REQ-PARTICIPANT")

    @property
    def participation_status(self) -> str:
        """Return the participation status, indicating whether the person will be present or not."""
        return self.get_property_parameter_default("PARTSTAT", default="NEEDS-ACTION")


class Attendee(_CalAddress):
    """The ATTENDEE property defines an "Attendee" within a calendar component."""

    pass


class Organizer(_CalAddress):
    """The ORGANIZER property defines the organizer for a calendar component."""

    pass


# if __name__ == "__main__":
#     ca = _CalAddress.create_property_from_str(None, "ORGANIZER;CN=John Smith:mailto:jsmith@example.com")
#     print(ca.email)
#     print(ca.persons_name)



from ical_library.base_classes.base_class import ICalBaseClass
from ical_library.base_classes.component import Component
from ical_library.ical_components import VCalendar
from ical_library.ical_properties.dt import RecurrenceID


def test_name(calendar_instance):
    assert ICalBaseClass(name="ABC", parent=calendar_instance).name == "ABC"


def test_parent(calendar_instance):
    another_component = Component(name="ANOTHER-COMPONENT", parent=calendar_instance)
    assert ICalBaseClass(name="ABC", parent=another_component).parent == another_component


def test_get_ical_name_of_class():
    class SomeRandomClass(ICalBaseClass):
        pass

    assert SomeRandomClass.get_ical_name_of_class() == "SOMERANDOMCLASS"

    class SomeOtherClass(ICalBaseClass):
        pass

    assert SomeOtherClass.get_ical_name_of_class() == "SOMEOTHERCLASS"
    assert VCalendar.get_ical_name_of_class() == "VCALENDAR"
    assert RecurrenceID.get_ical_name_of_class() == "RECURRENCE-ID"



import pendulum
from pendulum import DateTime

from ical_library.help_modules.timespan import Timespan
from ical_library.ical_components import VCalendar
from ical_library.timeline import Timeline


def test_vevent_with_exdate(calendar_exdate: VCalendar):
    timeline: Timeline = calendar_exdate.get_limited_timeline(
        DateTime(2022, 12, 27).in_tz("Europe/Amsterdam"), DateTime(2022, 12, 31).in_tz("Europe/Amsterdam")
    )
    all_events = list(timeline.iterate())
    assert len(all_events) == 2
    assert all_events[0][0] == Timespan(
        pendulum.parse("2022-12-27T11:15:00+01:00"), pendulum.parse("2022-12-27T11:45:00+01:00")
    )
    assert all_events[1][0] == Timespan(
        pendulum.parse("2022-12-29T11:15:00+01:00"), pendulum.parse("2022-12-29T11:45:00+01:00")
    )



from datetime import timedelta
from typing import Callable, Dict, Iterator, List, Optional, Tuple, Union

from pendulum import Date, DateTime
from pendulum.tz.timezone import Timezone
from pendulum.tz.zoneinfo.transition import Transition
from pendulum.tz.zoneinfo.transition_type import TransitionType

from ical_library.base_classes.component import Component
from ical_library.exceptions import MissingRequiredProperty
from ical_library.help_modules import dt_utils, property_utils
from ical_library.help_modules.lru_cache import instance_lru_cache
from ical_library.help_modules.timespan import Timespan
from ical_library.ical_properties.dt import DTStart, LastModified
from ical_library.ical_properties.pass_properties import Comment, TZID, TZName, TZURL
from ical_library.ical_properties.periods import RDate
from ical_library.ical_properties.rrule import RRule
from ical_library.ical_properties.tz_offset import TZOffsetFrom, TZOffsetTo


class _TimeOffsetPeriod(Component):
    """
    A _TimeOffsetPeriod representing either a Standard configuration or a Winter configuration.

    :param name: The actual name of this component instance. E.g. VEVENT, RRULE, VCUSTOMCOMPONENT.
    :param dtstart: The DTStart property. Required and must occur exactly once.
    :param tzoffsetto: The TZOffsetTo property. Required and must occur exactly once.
    :param tzoffsetfrom: The TZOffsetFrom property. Required and must occur exactly once.
    :param rrule: The RRule property. Optional, but may occur at most once.
    :param comment: The Comment property. Optional, but may occur multiple times.
    :param rdate: The RDate property. Optional, but may occur multiple times.
    :param tzname: The TZName property. Optional, but may occur multiple times.
    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    """

    def __init__(
        self,
        name: str,
        dtstart: Optional[DTStart] = None,
        tzoffsetto: Optional[TZOffsetTo] = None,
        tzoffsetfrom: Optional[TZOffsetFrom] = None,
        rrule: Optional[RRule] = None,
        comment: Optional[List[Comment]] = None,
        rdate: Optional[List[RDate]] = None,
        tzname: Optional[List[TZName]] = None,
        parent: Optional[Component] = None,
        is_dst: bool = False,
    ):
        super().__init__(name, parent=parent)
        # Required, must occur only once.
        self.dtstart: Optional[DTStart] = self.as_parent(dtstart)
        self.tzoffsetto: Optional[TZOffsetTo] = self.as_parent(tzoffsetto)
        self.tzoffsetfrom: Optional[TZOffsetFrom] = self.as_parent(tzoffsetfrom)
        # Optional, may only occur once.
        self.rrule: Optional[RRule] = self.as_parent(rrule)
        # Optional, may occur more than once.
        self.comment: Optional[List[Comment]] = self.as_parent(comment)
        self.rdate: Optional[List[RDate]] = self.as_parent(rdate)
        self.tzname: Optional[List[TZName]] = self.as_parent(tzname)
        self.is_dst = is_dst

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"{self.__class__.__name__}({self.dtstart.value}: {self.tzoffsetto.value})"

    def timezone_aware_start(self) -> DateTime:
        """Return a timezone aware start."""
        dt: DateTime = dt_utils.convert_time_object_to_datetime(self.dtstart.datetime_or_date_value)
        return dt.in_timezone(tz=self.tzoffsetfrom.as_timezone_object())

    def get_time_sequence(
        self, max_datetime: Optional[DateTime] = None
    ) -> Iterator[Tuple[DateTime, "_TimeOffsetPeriod"]]:
        """
        Expand the TimeZone start date according to its recurring *RDate* and *RRule* properties.
        :param max_datetime: The maximum datetime value we wish to expand to.
        :return: Yield all the datetime values according to the recurring properties that are lower than *max_datetime*.
        """
        for rtime in property_utils.expand_event_in_range_only_return_first(
            rdate_list=self.rdate or [],
            rrule=self.rrule,
            first_event_start=self.timezone_aware_start(),
            return_range=Timespan(self.timezone_aware_start() - timedelta(days=1), max_datetime),
            make_tz_aware=self.tzoffsetfrom.as_timezone_object(),
        ):
            if not isinstance(rtime, DateTime):
                raise TypeError(f"{rtime} was expected to be a DateTime object.")
            yield rtime, self


class DayLight(_TimeOffsetPeriod):
    """A TimeOffsetPeriod representing a DayLight(a.k.a. Advanced Time, Summer Time or Legal Time) configuration."""

    def __init__(self, parent: Optional[Component] = None, **kwargs):
        super().__init__("DAYLIGHT", parent=parent, is_dst=True, **kwargs)

    @classmethod
    def _get_init_method_for_var_mapping(cls) -> Callable:
        return _TimeOffsetPeriod.__init__


class Standard(_TimeOffsetPeriod):
    """A TimeOffsetPeriod representing a Standard(a.k.a. Winter Time) configuration."""

    def __init__(self, parent: Optional[Component] = None, **kwargs):
        super().__init__("STANDARD", parent=parent, is_dst=False, **kwargs)

    @classmethod
    def _get_init_method_for_var_mapping(cls) -> Callable:
        return _TimeOffsetPeriod.__init__


class VTimeZone(Component):
    """
    This class represents the VTIMEZONE component specified in RFC 5545 in '3.6.5. Time Zone Component'.

    If present, the "VTIMEZONE" calendar component defines the set of Standard Time and Daylight Saving Time
    observances (or rules) for a particular time zone for a given interval of time. The "VTIMEZONE" calendar component
    cannot be nested within other calendar components. Multiple "VTIMEZONE" calendar components can exist in an
    iCalendar object. In this situation, each "VTIMEZONE" MUST represent a unique time zone definition. This is
    necessary for some classes of events, such as airline flights, that start in one time zone and end in another.

    :param parent: The Component this item is encapsulated by in the iCalendar data file.
    :param tzid: The TZID property. Required and must occur exactly once.
    :param last_mod: Optional LastModified property. Optional, but may occur at most once.
    :param tzurl: The TZURL property. Optional, but may occur at most once.
    :param standardc: Optional list of Standard components. Each component may occur multiple times. Either standardc
    or daylightc must contain at least one TimeOffsetPeriod.
    :param daylightc: Optional list of DayLight components. Each component may occur multiple times. Either standardc
    or daylightc must contain at least one TimeOffsetPeriod.
    """

    def __init__(
        self,
        tzid: Optional[TZID] = None,
        last_mod: Optional[LastModified] = None,
        tzurl: Optional[TZURL] = None,
        standardc: Optional[List[Standard]] = None,
        daylightc: Optional[List[DayLight]] = None,
        parent: Optional[Component] = None,
    ):
        super().__init__("VTIMEZONE", parent=parent)

        # Required properties, must occur one.
        self._tzid: Optional[TZID] = self.as_parent(tzid)
        # Optional properties, may only occur once.
        self.last_mod: Optional[LastModified] = self.as_parent(last_mod)
        self.tzurl: Optional[TZURL] = self.as_parent(tzurl)
        # Either one of these components must have at least one record. May occur multiple times.
        self.standardc: List[Standard] = standardc or []
        self.daylightc: List[DayLight] = daylightc or []

        self.__storage_of_results: Dict[DateTime, List[Tuple[DateTime, _TimeOffsetPeriod]]] = {}

    def __repr__(self) -> str:
        """Overwrite the repr to create a better representation for the item."""
        return f"VTimeZone({self.tzid.value})"

    @property
    def tzid(self) -> TZID:
        """A getter to ensure the required property is set."""
        if self._tzid is None:
            raise MissingRequiredProperty(self, "tzid")
        return self._tzid

    @tzid.setter
    def tzid(self, value: TZID):
        """A setter to set the required property."""
        self._tzid = value

    def get_ordered_timezone_overview(self, max_datetime: DateTime) -> List[Tuple[DateTime, _TimeOffsetPeriod]]:
        """
        Expand all TimeOffsetPeriod configuration and return them in an ordered by time fashion.
        :param max_datetime: The maximum datetime value we wish to expand to.
        :return: A sorted list on datetime containing tuples of datetime and offset period where the datetime is
        lower than *max_datetime*.
        """
        if max_datetime in self.__storage_of_results.keys():
            return self.__storage_of_results[max_datetime]
        all_timezones: List[Tuple[Union[DateTime, Date], _TimeOffsetPeriod]] = []
        for a_standard in self.standardc:
            all_timezones.extend(a_standard.get_time_sequence(max_datetime=max_datetime))
        for a_daylight in self.daylightc:
            all_timezones.extend(a_daylight.get_time_sequence(max_datetime=max_datetime))
        sorted_list = sorted(all_timezones, key=lambda tup: tup[0])
        self.__storage_of_results[max_datetime] = sorted_list
        return sorted_list

    def convert_naive_datetime_to_aware(self, dt: DateTime) -> DateTime:
        """
        Convert a naive datetime to an aware datetime using the configuration of this TimeZone object.
        :param dt: The (possibly naive) datetime to convert to this timezone configuration.
        :return: The timezone aware datetime.
        """
        if dt.tzinfo is not None:
            return dt
        return dt.in_timezone(self.get_as_timezone_object())

    def get_ordered_timezone_overview_as_transition(self, max_datetime: DateTime) -> List[Transition]:
        """
        Get timezone components as a list of pendulum Transitions.
        :param max_datetime: The maximum datetime for which we include transitions. Any transitions after are excluded.
        :return: Return the list of pendulum Transitions for this VTimezone.
        """
        timezones: List[Tuple[DateTime, _TimeOffsetPeriod]] = self.get_ordered_timezone_overview(max_datetime)
        transitions: List[Transition] = []
        previous_transition: Optional[Transition] = None
        for time, offset_period in timezones:
            new_transition_type = TransitionType(
                offset=offset_period.tzoffsetto.parse_value_as_seconds(),
                is_dst=offset_period.is_dst,
                abbr=offset_period.tzname[0].value if offset_period.tzname else "unknown",
            )
            at = int(time.in_tz("UTC").timestamp())
            new_transition = Transition(at=at, ttype=new_transition_type, previous=previous_transition)
            previous_transition = new_transition
            transitions.append(new_transition)
        return transitions

    @instance_lru_cache()
    def get_as_timezone_object(self, max_datetime: DateTime = DateTime(2100, 1, 1)) -> Timezone:
        """
        For a given maximum datetime, compute a pendulum Timezone object that contains all transition till max_datetime.
        :param max_datetime: The maximum datetime for which we include transitions. Any transitions after are excluded.
        :return: Returns a pendulum Timezone object that you can use for DateTimes.
        """
        return CustomTimezone(
            name=self.tzid.value, transitions=self.get_ordered_timezone_overview_as_transition(max_datetime)
        )


class CustomTimezone(Timezone):
    def __init__(self, name: str, transitions: List[Transition]):  # noqa
        self._name = name
        self._transitions = transitions
        self._hint = {True: None, False: None}



import hashlib
import os
from pathlib import Path
from typing import Any, Union
from urllib import request

from pendulum import DateTime, Duration

from ical_library import client
from ical_library.ical_components import VCalendar


class CacheClient:
    """
    A iCalendar client which takes care of caching the result for you.
    This avoids you needing to handle the caching.

    :param cache_location: A path to the cache. Can be relative or absolute references. When you pass in a value with
    a file extension, it is considered to be a directory, otherwise it's considered as a file reference.
    :param cache_ttl: The time-to-live for the cache. The cache will be deleted/refreshed once it is older than the TTL.
    :param verbose: Print verbose messages regarding cache usage.
    :param url: The URL to the iCalendar file.
    """

    def __init__(
        self,
        url: str,
        cache_location: Union[Path, str],
        cache_ttl: Union[Duration] = Duration(hours=1),
        verbose: bool = True,
    ):
        self.url: str = url
        self.cache_location: Path = Path(cache_location)
        self.cache_ttl: Duration = cache_ttl
        self.verbose = verbose

    @property
    def cache_file_path(self) -> Path:
        """Return the filepath to the cache for the given URL."""
        if self.cache_location.suffix == "":
            return self.cache_location / hashlib.md5(self.url.encode()).hexdigest()
        return self.cache_location

    def get_icalendar(self, **kwargs: Any) -> VCalendar:
        """
        Get a parsed VCalendar instance. If there is an active cache, return that, otherwise fetch and cache the result.
        :param kwargs: Any keyword arguments to pass onto the `urllib.request.urlopen` call.
        :return: a VCalendar instance with all it's iCalendar components like VEvents, VToDos, VTimeZones etc.
        """
        if not self._is_cache_expired():
            if self.verbose:
                print("Using cache to remove this folder.")
            return client.parse_icalendar_file(self.cache_file_path)

        self._purge_icalendar_cache()
        response = request.urlopen(self.url, **kwargs)
        if not (200 <= response.getcode() < 400):
            raise ValueError(f"Unable to execute request at {self.url=}. Response code was: {response.getcode()}.")
        text = response.read().decode("utf-8")
        self._write_response_to_cache(text)

        lines = text.split("\n")
        return client.parse_lines_into_calendar(lines)

    def _write_response_to_cache(self, text: str) -> None:
        """
        Write the response of the fetched URL to cache.
        :param text: The fetched result.
        """
        if self.verbose:
            print(f"Successfully loaded new iCalendar data and stored it at {self.cache_file_path}.")
        self.cache_file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.cache_file_path, "w") as file:
            file.write(text)

    def _purge_icalendar_cache(self) -> None:
        """Purge the cache we have for this Calendar."""
        if self.verbose:
            print(f"Cache was expired. Removed {self.cache_file_path}.")
        return self.cache_file_path.unlink()

    def _is_cache_expired(self) -> bool:
        """Return whether the cache is passed its expiration date."""
        cutoff = DateTime.utcnow() - self.cache_ttl
        mtime = DateTime.utcfromtimestamp(os.path.getmtime(self.cache_file_path))
        return mtime < cutoff
